"""Shared constant definitions."""

import datetime as dt
from pathlib import Path
from platform import platform, python_version
from sys import argv, gettrace, version_info

VERSION: str

try:
    from mnamer.__version__ import __version__ as VERSION  # type: ignore
except ModuleNotFoundError:
    from setuptools_scm import get_version  # type: ignore

    VERSION = get_version(root="..", relative_to=__file__, local_scheme="dirty-tag")

try:
    from appdirs import __version__ as appdirs_version  # type: ignore
except ModuleNotFoundError:
    appdirs_version = "N/A"

try:
    from appdirs import user_cache_dir

    cache_dir = user_cache_dir()
except ModuleNotFoundError:
    cache_dir = "N/A"

try:
    from guessit import __version__ as guessit_version  # type: ignore
except ModuleNotFoundError:
    guessit_version = "N/A"

try:
    from requests import __version__ as requests_version
except ModuleNotFoundError:
    requests_version = "N/A"

try:
    from requests_cache import __version__ as requests_cache_version
except ModuleNotFoundError:
    requests_cache_version = "N/A"

try:
    from teletype import VERSION as teletype_version
except ModuleNotFoundError:
    teletype_version = "N/A"


CACHE_PATH = Path(
    cache_dir, f"mnamer-py{version_info.major}.{version_info.minor}"
).absolute()

CURRENT_YEAR = dt.datetime.now().year

DEPRECATED = {"no_replace", "replacements"}

IS_DEBUG = gettrace() is not None

SUBTITLE_CONTAINERS = [".srt", ".idx", ".sub"]


SYSTEM = {
    "date": dt.date.today(),
    "platform": platform(),
    "arguments": argv[1:],
    "cache location": f"{CACHE_PATH}.sqlite",
    "python version": python_version(),
    "mnamer version": VERSION,
    "appdirs version": appdirs_version,
    "guessit version": guessit_version,
    "requests version": requests_version,
    "requests cache version": requests_cache_version,
    "teletype version": teletype_version,
}

USAGE = "USAGE: mnamer [preferences] [directives] target [targets ...]"



import datetime as dt
from typing import Any, NamedTuple

from mnamer.const import SUBTITLE_CONTAINERS
from mnamer.language import Language
from mnamer.types import ProviderType

DEFAULT_SETTINGS = {
    "batch": False,
    "config_dump": False,
    "config_ignore": False,
    "episode_api": ProviderType.TVMAZE,
    "episode_directory": None,
    "episode_format": "{series} - S{season:02}E{episode:02} - {title}.{extension}",
    "hits": 5,
    "id_imdb": None,
    "id_tmdb": None,
    "id_tvdb": None,
    "id_tvmaze": None,
    "ignore": [".*sample.*", "^RARBG.*"],
    "lower": False,
    "mask": [".avi", ".m4v", ".mp4", ".mkv", ".ts", ".wmv"] + SUBTITLE_CONTAINERS,
    "media": None,
    "movie_api": ProviderType.TMDB,
    "movie_directory": None,
    "movie_format": "{name} ({year}).{extension}",
    "no_cache": False,
    "no_guess": False,
    "no_overwrite": False,
    "no_style": False,
    "recurse": False,
    "replace_after": {"&": "and", ";": ",", "@": "at"},
    "replace_before": {},
    "scene": False,
    "targets": [],
    "test": False,
    "verbose": False,
    "version": False,
}


JUNK_TEXT = "blablablabla"

EPISODE_META = {
    "The Walking Dead": {
        "date": dt.date(2015, 2, 22),
        "episode": 11,
        "id_imdb": "tt1520211",
        "id_tvdb": 153021,
        "id_tvmaze": 73,
        "media": "television",
        "season": 5,
        "series": "The Walking Dead",
        "title": "The Distance",
    },
    "Downtown": {
        "date": dt.date(1999, 11, 8),
        "episode": 13,
        "id_imdb": "tt0208616",
        "id_tvdb": 78342,
        "id_tvmaze": 30436,
        "media": "television",
        "season": 1,
        "series": "Downtown",
        "title": "Trip or Treat",
    },
    "Fargo": {
        "date": dt.date(2015, 10, 19),
        "episode": 2,
        "id_imdb": "tt2802850",
        "id_tvdb": 269613,
        "id_tvmaze": 32,
        "media": "television",
        "season": 2,
        "series": "Fargo",
        "title": "Before the Law",
    },
}

MOVIE_META = {
    "The Goonies": {
        "id_imdb": "tt0089218",
        "id_tmdb": 9340,
        "media": "movie",
        "name": "The Goonies",
        "year": 1985,
    },
    "Citizen Kane": {
        "id_imdb": "tt0033467",
        "id_tmdb": 15,
        "media": "movie",
        "name": "Citizen Kane",
        "year": 1941,
    },
    "Les Misérables": {
        "id_imdb": "tt10199590",
        "id_tmdb": 586863,
        "media": "movie",
        "name": "Les Misérables",
        "year": 2019,
    },
}

TEST_DATE = dt.date(2010, 12, 9)

RUSSIAN_LANG = Language.parse("ru")


class E2EResult(NamedTuple):
    code: int
    out: str


class MockRequestResponse:
    def __init__(self, status: int, content: str) -> None:
        self.status_code = status
        self.content = content

    def json(self) -> dict[str, Any]:
        from json import loads

        return loads(self.content)



from __future__ import annotations

import datetime as dt
from os import path
from pathlib import Path
from shutil import move
from typing import Any, ClassVar, Type

from guessit import guessit  # type: ignore

from mnamer.exceptions import MnamerException
from mnamer.language import Language
from mnamer.metadata import Metadata, MetadataEpisode, MetadataMovie
from mnamer.providers import Provider
from mnamer.setting_store import SettingStore
from mnamer.types import MediaType, ProviderType
from mnamer.utils import (
    crawl_in,
    filename_replace,
    filter_blacklist,
    filter_containers,
    is_subtitle,
    str_replace,
    str_sanitize,
    str_scenify,
)


class Target:
    """Manages metadata state for a media file and facilitates its relocation."""

    _providers: ClassVar[dict[ProviderType, Provider]] = {}

    _settings: SettingStore
    _provider: Provider
    _has_moved: bool
    _has_renamed: bool
    _raw_metadata: dict[str, str]
    _parsed_metadata: Metadata

    source: Path
    metadata: Metadata

    def __init__(self, file_path: Path, settings: SettingStore | None = None):
        self.source = file_path
        self._settings = settings or SettingStore()
        self._has_moved = False
        self._has_renamed = False
        self._parse(file_path)
        self._replace_before()
        self._override_metadata_ids()
        self._register_provider()

    def __str__(self) -> str:
        if isinstance(self.source, Path):
            return str(self.source.resolve())
        else:
            return str(self.source)

    @classmethod
    def populate_paths(cls: Type[Target], settings: SettingStore) -> list[Target]:
        """Creates a list of Target objects for media files found in paths."""
        file_paths = crawl_in(settings.targets, settings.recurse)
        file_paths = filter_blacklist(file_paths, settings.ignore)
        file_paths = filter_containers(file_paths, settings.mask)
        targets = [cls(file_path, settings) for file_path in file_paths]
        targets = list(dict.fromkeys(targets))  # unique values
        targets = list(filter(cls._matches_media, targets))
        return targets

    @classmethod
    def reset_providers(cls):
        cls._providers.clear()

    @staticmethod
    def _matches_media(target: Target) -> bool:
        if not target._settings.media:
            return True
        else:
            return target._settings.media is target.metadata.to_media_type()

    @property
    def provider_type(self) -> ProviderType:
        provider_type = self._settings.api_for(self.metadata.to_media_type())
        assert provider_type
        return provider_type

    @property
    def directory(self) -> Path | None:
        settings_key = f"{self.metadata.to_media_type().value}_directory"
        directory = getattr(self._settings, settings_key)
        return Path(directory) if directory else None

    @property
    def destination(self) -> Path:
        """
        The destination Path for the target based on its metadata and user
        preferences.
        """
        if self.directory:
            dir_head_ = format(self.metadata, str(self.directory))
            dir_head_ = str_sanitize(dir_head_)
            dir_head = Path(dir_head_)
        else:
            dir_head = self.source.parent
        file_path = format(self.metadata, self._settings.formatting_for(self.metadata))
        dir_tail, filename = path.split(Path(file_path))
        filename = filename_replace(filename, self._settings.replace_after)
        if self._settings.scene:
            filename = str_scenify(filename)
        if self._settings.lower:
            filename = filename.lower()
        filename = str_sanitize(filename)
        directory = Path(dir_head, dir_tail)
        return Path(directory, filename)

    def _parse(self, file_path: Path):
        path_data: dict[str, Any] = {"language": self._settings.language}
        if is_subtitle(self.source):
            try:
                path_data["language"] = Language.parse(self.source.stem[-2:])
                file_path = Path(self.source.parent, self.source.stem[:-2])
            except MnamerException:
                pass
        options = {"type": self._settings.media, "language": path_data["language"]}
        raw_data = dict(guessit(str(file_path), options))
        if isinstance(raw_data.get("season"), list):
            raw_data = dict(guessit(str(file_path.parts[-1]), options))
        for k, v in raw_data.items():
            if hasattr(v, "alpha3"):
                try:
                    path_data[k] = Language.parse(v)
                except MnamerException:
                    continue
            elif isinstance(v, (int, str, dt.date)):
                path_data[k] = v
            elif isinstance(v, list) and all([isinstance(_, (int, str)) for _ in v]):
                path_data[k] = v[0]
        if self._settings.media:
            media_type = self._settings.media
        elif path_data.get("type"):
            media_type = MediaType(path_data["type"])
        else:
            media_type = None
        meta_cls = {
            MediaType.EPISODE: MetadataEpisode,
            MediaType.MOVIE: MetadataMovie,
            None: Metadata,
        }[media_type]
        self.metadata = meta_cls()
        self.metadata.quality = (
            " ".join(
                path_data[key]
                for key in path_data
                if key
                in (
                    "audio_codec",
                    "audio_profile",
                    "screen_size",
                    "source",
                    "video_codec",
                    "video_profile",
                )
            )
            or None
        )
        self.metadata.language = path_data.get("language")
        self.metadata.group = path_data.get("release_group")
        self.metadata.container = file_path.suffix or None
        if not self.metadata.language:
            try:
                self.metadata.language = path_data.get("language")
            except MnamerException:
                pass
        try:
            self.metadata.language_sub = path_data.get("subtitle_language")
        except MnamerException:
            pass
        if isinstance(self.metadata, MetadataMovie):
            self.metadata.name = path_data.get("title")
            self.metadata.year = path_data.get("year")
        elif isinstance(self.metadata, MetadataEpisode):
            self.metadata.date = path_data.get("date")
            self.metadata.episode = path_data.get("episode")
            self.metadata.season = path_data.get("season")
            self.metadata.series = path_data.get("title")
            alternative_title = path_data.get("alternative_title")
            if alternative_title:
                self.metadata.series = f"{self.metadata.series} {alternative_title}"
            # adding year to title can reduce false positives
            # year = path_data.get("year")
            # if year:
            #     self.metadata.series = f"{self.metadata.series} {year}"

    def _override_metadata_ids(self):
        id_types = {"imdb", "tmdb", "tvdb", "tvmaze"}
        for id_type in id_types:
            attr = f"id_{id_type}"
            if not hasattr(self.metadata, attr):
                continue  # ensure metadata subclass supports id type
            value = getattr(self._settings, attr, None)
            if not value:
                continue  # apply override if set in directives
            setattr(self.metadata, attr, value)

    def _register_provider(self) -> None:
        provider_type = self.provider_type
        if provider_type and provider_type not in self._providers:
            self._providers[provider_type] = Provider.provider_factory(
                provider_type, self._settings
            )
        self._provider = self._providers[provider_type]

    def _replace_before(self) -> None:
        if not self._settings.replace_before:
            return
        for attr, value in vars(self.metadata).items():
            if not isinstance(value, str):
                continue
            if attr.startswith("_"):
                continue
            value = str_replace(value, self._settings.replace_before)
            setattr(self.metadata, attr, value)

    def query(self) -> list[Metadata]:
        """Queries the target's respective media provider for metadata."""
        results = self._provider.search(self.metadata)
        if not results:
            return []
        seen = set()
        response = []
        for idx, result in enumerate(results, start=1):
            if str(result) in seen:
                continue
            response.append(result)
            seen.add(str(result))
            if idx >= self._settings.hits:
                break
        return response

    def relocate(self) -> None:
        """Performs the action of renaming and/or moving a file."""
        destination_path = Path(self.destination).resolve()
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            move(str(self.source), destination_path)
        except OSError:  # pragma: no cover
            raise MnamerException



import os
import tempfile
from pathlib import Path
from shutil import rmtree

import pytest


@pytest.fixture
def setup_test_dir(request):
    orig_dir = os.getcwd()
    tmp_dir = tempfile.mkdtemp()
    os.chdir(tmp_dir)

    def finalizer():
        os.chdir(orig_dir)
        rmtree(tmp_dir)

    request.addfinalizer(finalizer)


@pytest.fixture
def setup_test_files():
    def fn(*filenames):
        paths = [Path(*filename.split("/")) for filename in filenames]
        for path in paths:
            path.parent.mkdir(parents=True, exist_ok=True)
            path.open("w").close()

    return fn



"""Provides an interface for handling user input and printing output."""

import traceback
from typing import Any, Callable

from teletype import codes
from teletype.components import ChoiceHelper, SelectOne
from teletype.io import style_format, style_print

from mnamer.const import SYSTEM
from mnamer.exceptions import MnamerAbortException, MnamerException, MnamerSkipException
from mnamer.language import Language
from mnamer.metadata import Metadata
from mnamer.setting_store import SettingStore
from mnamer.types import MessageType
from mnamer.utils import format_dict, format_exception, format_iter

no_style: bool = False
verbose: bool = False


def _chars() -> dict[str, str]:
    if no_style:
        chars = codes.CHARS_ASCII
    else:
        chars = codes.CHARS_DEFAULT
        chars["arrow"] = style_format(chars["arrow"], "magenta")
    return chars


def _abort_helpers() -> tuple[
    ChoiceHelper[MnamerSkipException], ChoiceHelper[MnamerAbortException]
]:
    if no_style:
        style = None
        skip_mnemonic = "[s]"
        quit_mnemonic = "[q]"
    else:
        style = "dark"
        skip_mnemonic = "s"
        quit_mnemonic = "q"
    return (
        ChoiceHelper(MnamerSkipException(), "skip", style, skip_mnemonic),
        ChoiceHelper(MnamerAbortException(), "quit", style, quit_mnemonic),
    )


def _msg_format(body: Any):
    converter_map: dict[type, Callable] = {
        dict: format_dict,
        list: format_iter,
        tuple: format_iter,
        set: format_iter,
        MnamerException: format_exception,
    }
    converter: Callable | None = converter_map.get(type(body), str)
    if converter:
        body = converter(body)
    else:
        body = getattr(body, "value", body)
    return body


def configure(settings: SettingStore):
    """Sets class variables using a settings instance."""
    global verbose, no_style
    verbose = settings.verbose
    no_style = settings.no_style


def msg(
    body: Any,
    message_type: MessageType = MessageType.INFO,
    debug: bool = False,
):
    if debug and not verbose:
        return
    if no_style:
        print(_msg_format(body))
    else:
        style_print(_msg_format(body), style=message_type.value)


def error(body: Any):
    msg(body, message_type=MessageType.ERROR, debug=False)


def metadata_prompt(matches: list[Metadata]) -> Metadata:  # pragma: no cover
    """Prompts user to choose a match from a list of matches."""
    msg("select match")
    selector = SelectOne([*matches, *_abort_helpers()], **_chars())
    choice = selector.prompt()
    if isinstance(choice, (MnamerAbortException, MnamerSkipException)):
        raise choice
    return choice


def metadata_guess(metadata: Metadata) -> Metadata:  # pragma: no cover
    """Prompts user to confirm a single match."""
    label = str(metadata)
    if no_style:
        label += " (best guess)"
    else:
        label += style_format(" (best guess)", "blue")
    option = ChoiceHelper(metadata, label)
    selector = SelectOne([option, *_abort_helpers()], **_chars())
    choice = selector.prompt()
    if isinstance(choice, (MnamerAbortException, MnamerSkipException)):
        raise choice
    else:
        return choice


def subtitle_prompt() -> Language:
    msg("select language")
    choices = [ChoiceHelper(language, language.name) for language in Language.all()]
    selector = SelectOne([*choices, *_abort_helpers()], **_chars())
    choice = selector.prompt()
    if isinstance(choice, (MnamerAbortException, MnamerSkipException)):
        raise choice
    else:
        return choice


def crash_report():  # pragma: no cover
    s = f"""
============================== CRASH REPORT BEGIN ==============================

--------------------------------- environment ----------------------------------

{_msg_format(SYSTEM)}

--------------------------------- stack trace ----------------------------------

{traceback.format_exc()}
=============================== CRASH REPORT END ===============================

Dang, it looks like mnamer crashed! Please consider filling an issue at
https://github.com/jkwill87/mnamer/issues along with this report.
"""
    print(s)
    raise SystemExit(1)



import datetime as dt

import pytest

from mnamer.exceptions import MnamerNotFoundException
from mnamer.metadata import MetadataEpisode
from mnamer.providers import Tvdb
from tests import EPISODE_META, JUNK_TEXT

pytestmark = [
    pytest.mark.network,
    pytest.mark.tvdb,
    pytest.mark.flaky(reruns=5, reruns_delay=2),
]


@pytest.fixture(scope="session")
def provider():
    return Tvdb()


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id(meta: dict, provider: Tvdb):
    query = MetadataEpisode(id_tvdb=meta["id_tvdb"])
    results = list(provider.search(query))
    assert results
    for result in results:
        assert result.series == meta["series"]


def test_search_id__no_hits(provider: Tvdb):
    query = MetadataEpisode(id_tvdb=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search__series(meta: dict, provider: Tvdb):
    query = MetadataEpisode(series=meta["series"])
    assert any(result.id_tvdb == meta["id_tvdb"] for result in provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search__id_tvdb_season(meta: dict, provider: Tvdb):
    query = MetadataEpisode(season=1, id_tvdb=meta["id_tvdb"])
    results = provider.search(query)
    all_season_1 = all(entry.season == 1 for entry in results)
    assert all_season_1 is True


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search__id_tvdb_episode(meta: dict, provider: Tvdb):
    query = MetadataEpisode(episode=2, id_tvdb=meta["id_tvdb"])
    results = provider.search(query)
    all_episode_2 = all(entry.episode == 2 for entry in results)
    assert all_episode_2 is True


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_tvdb_provider__search__id_tvdb_season_episode(meta: dict, provider: Tvdb):
    query = MetadataEpisode(season=1, episode=3, id_tvdb=meta["id_tvdb"])
    results = list(provider.search(query))
    assert len(results) == 1
    assert results[0].season == 1
    assert results[0].episode == 3


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_tvdb_provider__search__series(meta: dict, provider: Tvdb):
    query = MetadataEpisode(series=meta["series"])
    found = False
    results = provider.search(query)
    for result in results:
        if result.id_tvdb == meta["id_tvdb"]:
            found = True
            break
    assert found is True


def test_tvdb_provider__search__series_deep(provider: Tvdb):
    query = MetadataEpisode(series="House Rules (au)", season=6, episode=6)
    results = provider.search(query)
    assert any(result.id_tvdb == 269795 for result in results)


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_tvdb_provider__search__title_season(meta: dict, provider: Tvdb):
    query = MetadataEpisode(series=meta["series"], season=1)
    results = provider.search(query)
    all_season_1 = all(entry.season == 1 for entry in results)
    assert all_season_1 is True


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_tvdb_provider__search__title_season_episode(meta: dict, provider: Tvdb):
    query = MetadataEpisode(series=meta["series"], season=1, episode=3)
    results = list(provider.search(query))
    assert results[0].season == 1
    assert results[0].episode == 3


def test_tvdb_provider__search_series_date__year(provider: Tvdb):
    query = MetadataEpisode(series="The Daily Show", date=dt.date(2017, 11, 1))
    results = list(provider.search(query))
    assert len(results) == 1
    assert results[0].title == "Hillary Clinton"



from __future__ import annotations

import dataclasses
import datetime as dt
import re
from string import Formatter
from typing import Any, Callable, Mapping, Sequence

from mnamer.language import Language
from mnamer.types import MediaType
from mnamer.utils import (
    fn_pipe,
    is_subtitle,
    normalize_container,
    parse_date,
    str_fix_padding,
    str_replace_slashes,
    str_title_case,
    year_parse,
)


class _MetaFormatter(Formatter):
    def format_field(self, value: None | int | str, format_spec: str) -> str:
        return format(value, format_spec) if value is not None else ""

    def get_value(
        self, key: str | int, args: Sequence[Any], kwargs: Mapping[str, Any]
    ) -> None | int | str:
        if isinstance(key, int):
            assert args
            return args[key]
        else:
            return kwargs.get(key, "")


@dataclasses.dataclass
class Metadata:
    """A dataclass which transforms and stores media metadata information."""

    container: str | None = None
    group: str | None = None
    language: Language | None = None
    language_sub: Language | None = None
    quality: str | None = None
    synopsis: str | None = None

    @classmethod
    def to_media_type(cls) -> MediaType:
        if cls is MetadataEpisode:
            return MediaType.EPISODE
        elif cls is MetadataMovie:
            return MediaType.MOVIE
        else:
            raise ValueError(f"Unknown metadata class: {cls}")

    def __setattr__(self, key: str, value: Any):
        converter_map: dict[str, Callable] = {
            "container": normalize_container,
            "group": str.upper,
            "language": Language.parse,
            "language_sub": Language.parse,
            "media": MediaType,
            "quality": str.lower,
            "synopsis": str.capitalize,
        }
        converter: Callable | None = converter_map.get(key)
        if value is not None and converter:
            value = converter(value)
        super().__setattr__(key, value)

    def __format__(self, format_spec: str | None):
        raise NotImplementedError

    def __str__(self) -> str:
        return self.__format__(None)

    @property
    def extension(self):
        if is_subtitle(self.container) and self.language_sub:
            return f".{self.language_sub.a2}{self.container}"
        else:
            return self.container

    def as_dict(self) -> dict[str, Any]:
        d = dataclasses.asdict(self)
        d["extension"] = self.extension
        return d

    def _format_repl(self, mobj) -> str:
        format_string, key = mobj.groups()
        value = _MetaFormatter().vformat(format_string, "", self.as_dict())
        if key in {"name", "series", "synopsis", "title"}:
            value = str_title_case(value)
        return value

    def update(self, metadata: Metadata):
        """Overlays all none value from another Metadata instance."""
        for field in dataclasses.asdict(self).keys():
            value = getattr(metadata, field)
            if value is None:
                continue
            super().__setattr__(field, value)


@dataclasses.dataclass
class MetadataMovie(Metadata):
    """
    A dataclass which transforms and stores media metadata information specific
    to movies.
    """

    name: str | None = None
    year: str | None = None
    id_imdb: str | None = None
    id_tmdb: str | None = None

    def __format__(self, format_spec: str | None):
        default = "{name} ({year})"
        re_pattern = r"({(\w+)(?:\[[\w:]+\])?(?:\:\d{1,2})?})"
        s = re.sub(re_pattern, self._format_repl, format_spec or default)
        s = str_fix_padding(s)
        return s

    def __setattr__(self, key: str, value: Any):
        converter_map: dict[str, Callable] = {
            "name": fn_pipe(str_replace_slashes, str_title_case),
            "year": year_parse,
        }
        converter: Callable | None = converter_map.get(key)
        if value is not None and converter:
            value = converter(value)
        super().__setattr__(key, value)


@dataclasses.dataclass
class MetadataEpisode(Metadata):
    """
    A dataclass which transforms and stores media metadata information specific to
    television episodes.
    """

    series: str | None = None
    season: int | None = None
    episode: int | None = None
    date: dt.date | None = None
    title: str | None = None
    id_tvdb: str | None = None
    id_tvmaze: str | None = None

    def __post_init__(self):
        if isinstance(self.season, str):
            self.season = int(self.season)
        if isinstance(self.episode, str):
            self.episode = int(self.episode)
        if isinstance(self.date, str):
            self.date = parse_date(self.date)

    def __format__(self, format_spec: str | None):
        default = "{series} - {season:02}x{episode:02} - {title}"
        re_pattern = r"({(\w+)(?:\[[\w:]+\])?(?:\:\d{1,2})?})"
        s = re.sub(re_pattern, self._format_repl, format_spec or default)
        s = str_fix_padding(s)
        return s

    def __setattr__(self, key: str, value: Any):
        converter_map: dict[str, Callable] = {
            "date": parse_date,
            "episode": int,
            "season": int,
            "series": fn_pipe(str_replace_slashes, str_title_case),
            "title": fn_pipe(str_replace_slashes, str_title_case),
        }
        converter: Callable | None = converter_map.get(key)
        if value is not None and converter:
            value = converter(value)
        super().__setattr__(key, value)



import pytest

from mnamer.setting_spec import SettingSpec
from mnamer.types import SettingType

pytestmark = pytest.mark.local


def test_setting_spec__serialize__default():
    default = {
        "action": None,
        "choices": None,
        "dest": None,
        "flags": None,
        "group": SettingType.PARAMETER,
        "help": None,
        "nargs": None,
        "typevar": None,
    }
    setting_spec = SettingSpec(SettingType.PARAMETER)
    assert setting_spec.as_dict() == default


def test_setting_spec__serialize__override():
    spec = {
        "action": "count",
        "choices": ["a", "b", "c"],
        "dest": "foo",
        "flags": ["-f", "--foo"],
        "group": SettingType.PARAMETER,
        "help": "foos your bars",
        "nargs": "+",
        "typevar": int,
    }
    setting_spec = SettingSpec(**spec)
    assert setting_spec.as_dict() == spec


def test_setting_spec__registration():
    spec = {
        "dest": "foo",
        "flags": ["-f", "--f"],
        "group": SettingType.DIRECTIVE,
        "help": "foos your bars",
    }
    setting_spec = SettingSpec(**spec)
    args, kwargs = setting_spec.registration
    assert args == ["-f", "--f"]
    assert kwargs == {"dest": "foo", "help": "foos your bars"}


def test_setting_spec__name__flags():
    spec = {
        "dest": "foo",
        "flags": ["-f", "--foo"],
        "group": SettingType.DIRECTIVE,
        "help": "foos your bars",
    }
    setting_spec = SettingSpec(**spec)
    assert setting_spec.name == "foo"


def test_setting_spec__name__no_flags():
    spec = {
        "dest": "foo",
        "group": SettingType.CONFIGURATION,
        "help": "foos your bars",
    }
    setting_spec = SettingSpec(**spec)
    assert setting_spec.name is None



import pytest

from mnamer.exceptions import MnamerNotFoundException
from mnamer.metadata import MetadataMovie
from mnamer.providers import Tmdb
from tests import JUNK_TEXT, MOVIE_META

pytestmark = [
    pytest.mark.network,
    pytest.mark.tmdb,
    pytest.mark.flaky(reruns=1),
]


@pytest.fixture(scope="session")
def provider():
    return Tmdb()


@pytest.mark.parametrize("meta", MOVIE_META.values(), ids=list(MOVIE_META))
def test_search_id(meta: dict, provider: Tmdb):
    query = MetadataMovie(id_tmdb=meta["id_tmdb"])
    results = list(provider.search(query))
    assert len(results) == 1
    result = results[0]
    assert result.name == meta["name"]


def test_search_id__no_hits(provider: Tmdb):
    query = MetadataMovie(id_tmdb=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", MOVIE_META.values(), ids=list(MOVIE_META))
def test_search__name(meta: dict, provider: Tmdb):
    query = MetadataMovie(name=meta["name"])
    assert any(result.id_tmdb == meta["id_tmdb"] for result in provider.search(query))


def test_search__no_hits(provider: Tmdb):
    query = MetadataMovie(name=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


def test_search__missing(provider: Tmdb):
    query = MetadataMovie()
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))



import pytest
from babelfish import Language as BabelLang  # type: ignore

from mnamer.language import Language


@pytest.mark.parametrize("value", ("english", "en", "eng"))
def test_language_parse__str(value):
    expected = Language("English", "en", "eng")
    actual = Language.parse(value)
    assert actual == expected


def test_language_parse__bl():
    expected = Language("English", "en", "eng")
    actual = Language.parse(BabelLang("eng"))
    assert actual == expected


def test_language_all():
    expected = (
        Language("Arabic", "ar", "ara"),
        Language("Chinese", "zh", "zho"),
        Language("Croatian", "hr", "hrv"),
        Language("Czech", "cs", "ces"),
        Language("Danish", "da", "dan"),
        Language("English", "en", "eng"),
        Language("French", "fr", "fra"),
        Language("German", "de", "deu"),
        Language("Greek", "el", "ell"),
        Language("Hebrew", "he", "heb"),
        Language("Hindi", "hi", "hin"),
        Language("Italian", "it", "ita"),
        Language("Japanese", "ja", "jpn"),
        Language("Korean", "ko", "kor"),
        Language("Latin", "la", "lat"),
        Language("Persian", "fa", "fas"),
        Language("Portuguese", "pt", "por"),
        Language("Russian", "ru", "rus"),
        Language("Slovenian", "sl", "slv"),
        Language("Spanish", "es", "spa"),
        Language("Swedish", "sv", "swe"),
        Language("Turkish", "tr", "tur"),
        Language("Ukrainian", "uk", "ukr"),
    )
    actual = Language.all()
    assert actual == expected



import pytest

from mnamer.endpoints import tmdb_find, tmdb_movies, tmdb_search_movies
from mnamer.exceptions import MnamerException, MnamerNotFoundException
from mnamer.providers import Tmdb
from tests import JUNK_TEXT, RUSSIAN_LANG

pytestmark = [
    pytest.mark.network,
    pytest.mark.tmdb,
    pytest.mark.flaky(reruns=1),
]

GOONIES_IMDB_ID = "tt0089218"
GOONIES_TMDB_ID = "9340"
JUNK_IMDB_ID = "tt1234567890"


def test_tmdb_find__imdb_success():
    expected_top_level_keys = {
        "movie_results",
        "person_results",
        "tv_episode_results",
        "tv_results",
        "tv_season_results",
    }
    expected_movie_results_keys = {
        "adult",
        "backdrop_path",
        "genre_ids",
        "id",
        "original_language",
        "original_title",
        "overview",
        "media_type",
        "poster_path",
        "popularity",
        "release_date",
        "title",
        "video",
        "vote_average",
        "vote_count",
    }
    result = tmdb_find(Tmdb.api_key, "imdb_id", GOONIES_IMDB_ID)
    assert isinstance(result, dict)
    assert set(result.keys()).issuperset(expected_top_level_keys)
    assert len(result.get("movie_results", {})) > 0
    assert expected_movie_results_keys.issuperset(
        set(result.get("movie_results", {})[0].keys())
    )


def test_tmdb_find__api_key_fail():
    with pytest.raises(MnamerException):
        tmdb_find(JUNK_TEXT, "imdb_id", GOONIES_IMDB_ID, cache=False)


def test_tmdb_find__invalid_id_imdb():
    with pytest.raises(MnamerException):
        tmdb_find(Tmdb.api_key, "imdb_id", JUNK_TEXT, cache=False)


def test_tmdb_find__not_found():
    with pytest.raises(MnamerNotFoundException):
        tmdb_find(Tmdb.api_key, "imdb_id", JUNK_IMDB_ID)


def test_tmdb_find__language():
    results = tmdb_find(Tmdb.api_key, "imdb_id", GOONIES_IMDB_ID, RUSSIAN_LANG)
    assert any(result["title"] == "Балбесы" for result in results["movie_results"])


def test_tmdb_find__invalid_source():
    with pytest.raises(MnamerException):
        tmdb_find(Tmdb.api_key, "abc123", GOONIES_IMDB_ID, cache=False)


def test_tmdb_movies__success():
    expected_top_level_keys = {
        "adult",
        "backdrop_path",
        "belongs_to_collection",
        "budget",
        "genres",
        "homepage",
        "id",
        "imdb_id",
        "original_language",
        "original_title",
        "overview",
        "popularity",
        "poster_path",
        "production_companies",
        "production_countries",
        "release_date",
        "revenue",
        "runtime",
        "spoken_languages",
        "status",
        "tagline",
        "title",
        "video",
        "vote_average",
        "vote_count",
    }
    result = tmdb_movies(Tmdb.api_key, GOONIES_TMDB_ID)
    assert isinstance(result, dict)
    assert set(result.keys()).issuperset(expected_top_level_keys)
    assert result.get("title") == "The Goonies"


def test_tmdb_movies__api_key_fail():
    with pytest.raises(MnamerException):
        tmdb_movies(JUNK_TEXT, "", cache=False)


def test_tmdb_movies__id_tmdb_fail():
    with pytest.raises(MnamerException):
        tmdb_movies(Tmdb.api_key, JUNK_TEXT, cache=False)


def test_tmdb_movies__not_found():
    with pytest.raises(MnamerNotFoundException):
        tmdb_movies(Tmdb.api_key, "1" * 10)


@pytest.mark.xfail(strict=False)
def test_tmdb_movies__language():
    result = tmdb_movies(Tmdb.api_key, GOONIES_TMDB_ID, RUSSIAN_LANG)
    assert result.get("title") == "Балбесы"


def test_tmdb_search_movies__success():
    expected_top_level_keys = {
        "page",
        "results",
        "total_pages",
        "total_results",
    }
    expected_results_keys = {
        "adult",
        "backdrop_path",
        "genre_ids",
        "id",
        "original_language",
        "original_title",
        "overview",
        "popularity",
        "poster_path",
        "release_date",
        "title",
        "video",
        "vote_average",
        "vote_count",
    }
    result = tmdb_search_movies(Tmdb.api_key, "the goonies", 1985)
    assert isinstance(result, dict)
    assert set(result.keys()).issuperset(expected_top_level_keys)
    assert isinstance(result["results"], list)
    assert expected_results_keys.issuperset(set(result.get("results", [{}])[0].keys()))
    assert result["results"][0]["original_title"] == "The Goonies"
    result = tmdb_search_movies(Tmdb.api_key, "the goonies")
    assert len(result["results"]) > 1


def test_tmdb_search_movies__bad_api_key():
    with pytest.raises(MnamerException):
        tmdb_search_movies(JUNK_TEXT, "the goonies", cache=False)


def test_tmdb_search_movies__bad_title():
    with pytest.raises(MnamerNotFoundException):
        tmdb_search_movies(Tmdb.api_key, JUNK_TEXT, cache=False)


def test_search_movies__language():
    results = tmdb_search_movies(Tmdb.api_key, "the goonies", language=RUSSIAN_LANG)
    assert any(result["title"] == "Балбесы" for result in results["results"])



from sys import platform

import pytest

from mnamer import tty
from mnamer.exceptions import MnamerAbortException, MnamerSkipException

pytestmark = pytest.mark.local


def test_chars():
    bullet = "►" if platform.startswith("win") else "❱"
    tty.verbose = False
    tty.no_style = False
    expected = {
        "arrow": f"\x1b[35m{bullet}\x1b[0m",
        "block": "█",
        "left-edge": "▐",
        "right-edge": "▌",
        "selected": "●",
        "unselected": "○",
    }
    actual = tty._chars()
    assert actual == expected


def test_chars__no_style():
    tty.verbose = False
    tty.no_style = True
    expected = {
        "arrow": ">",
        "block": "#",
        "left-edge": "|",
        "right-edge": "|",
        "selected": "*",
        "unselected": ".",
    }
    actual = tty._chars()
    assert actual == expected


def test_abort_helpers():
    tty.verbose = False
    tty.no_style = False
    helpers = tty._abort_helpers()
    assert len(helpers) == 2
    assert helpers[0].label == "skip"
    assert isinstance(helpers[0].value, MnamerSkipException)
    assert helpers[0]._bracketed is False
    assert helpers[1].label == "quit"
    assert isinstance(helpers[1].value, MnamerAbortException)
    assert helpers[1]._bracketed is False


def test_abort_helpers__no_style():
    tty.verbose = False
    tty.no_style = True
    helpers = tty._abort_helpers()
    assert len(helpers) == 2
    assert helpers[0].label == "skip"
    assert isinstance(helpers[0].value, MnamerSkipException)
    assert helpers[0]._bracketed is True
    assert helpers[1].label == "quit"
    assert isinstance(helpers[1].value, MnamerAbortException)
    assert helpers[1]._bracketed is True



import pytest

from mnamer.endpoints import (
    tvdb_episodes_id,
    tvdb_login,
    tvdb_refresh_token,
    tvdb_search_series,
    tvdb_series_id,
    tvdb_series_id_episodes,
    tvdb_series_id_episodes_query,
)
from mnamer.exceptions import MnamerException, MnamerNotFoundException
from mnamer.language import Language
from mnamer.providers import Tvdb
from tests import JUNK_TEXT, RUSSIAN_LANG

pytestmark = [
    pytest.mark.network,
    pytest.mark.tvdb,
    pytest.mark.flaky(reruns=1),
]

EXPECTED_TOP_LEVEL_SHOW_KEYS = {
    "absoluteNumber",
    "airedEpisodeNumber",
    "airedSeason",
    "airedSeasonID",
    "airsAfterSeason",
    "airsBeforeEpisode",
    "airsBeforeSeason",
    "contentRating",
    "directors",
    "dvdChapter",
    "dvdDiscid",
    "dvdEpisodeNumber",
    "dvdSeason",
    "episodeName",
    "filename",
    "firstAired",
    "guestStars",
    "id",
    "imdbId",
    "isMovie",
    "language",
    "lastUpdated",
    "lastUpdatedBy",
    "overview",
    "productionCode",
    "seriesId",
    "showUrl",
    "siteRating",
    "siteRatingCount",
    "thumbAdded",
    "thumbAuthor",
    "thumbHeight",
    "thumbWidth",
    "writers",
}


LOST_TVDB_ID_EPISODE = "127131"
LOST_TVDB_ID_SERIES = "73739"
THE_WITCHER_ID_SERIES = "362696"


@pytest.fixture(scope="session")
def tvdb_token():
    """Calls mnamer.endpoints.tvdb_login then returns cached token."""
    if not hasattr(tvdb_token, "token"):
        from mnamer.endpoints import tvdb_login

        tvdb_token.token = tvdb_login(Tvdb.api_key)
    return tvdb_token.token


def test_tvdb_login__login_success():
    assert tvdb_login(Tvdb.api_key) is not None


def test_tvdb_login__login_fail():
    with pytest.raises(MnamerException):
        tvdb_login(JUNK_TEXT)


def test_tvdb_refresh_token__refresh_success():
    token = tvdb_login(Tvdb.api_key)
    assert tvdb_refresh_token(token) is not None


def test_tvdb_refresh_token__refresh_fail():
    with pytest.raises(MnamerException):
        tvdb_refresh_token(JUNK_TEXT)


@pytest.mark.xfail(strict=False)
def test_tvdb_episodes_id__invalid_token():
    with pytest.raises(MnamerException):
        tvdb_episodes_id(JUNK_TEXT, LOST_TVDB_ID_EPISODE, cache=False)


def test_tvdb_episodes_id__invalid_lang(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_episodes_id(
            tvdb_token,
            LOST_TVDB_ID_EPISODE,
            language=Language(JUNK_TEXT, JUNK_TEXT, JUNK_TEXT),
            cache=False,
        )


def test_tvdb_episodes_id__invalid_id_imdb(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_episodes_id(tvdb_token, JUNK_TEXT, cache=False)


def test_tvdb_episodes_id__no_hits(tvdb_token):
    with pytest.raises(MnamerNotFoundException):
        tvdb_episodes_id(tvdb_token, LOST_TVDB_ID_EPISODE * 2, cache=False)


def test_tvdb_episodes_id__success(tvdb_token):
    result = tvdb_episodes_id(tvdb_token, LOST_TVDB_ID_EPISODE)
    assert isinstance(result, dict)
    assert "data" in result
    assert set(result["data"].keys()) == EXPECTED_TOP_LEVEL_SHOW_KEYS
    assert str(result["data"]["seriesId"]) == LOST_TVDB_ID_SERIES
    assert str(result["data"]["id"]) == LOST_TVDB_ID_EPISODE


def test_tvdb_episodes_id__language(tvdb_token):
    result = tvdb_episodes_id(tvdb_token, LOST_TVDB_ID_EPISODE, RUSSIAN_LANG)
    assert result["data"]["episodeName"] == "Пилот (1)"


def test_tvdb_episodes_id__language__invalid(tvdb_token):
    invalid_language = Language("invalid", "xy", "xyz")
    with pytest.raises(MnamerException):
        tvdb_episodes_id(tvdb_token, LOST_TVDB_ID_EPISODE, invalid_language)


@pytest.mark.xfail(strict=False)
def test_tvdb_series_id__invalid_token():
    with pytest.raises(MnamerException):
        tvdb_series_id(JUNK_TEXT, LOST_TVDB_ID_SERIES, cache=False)


def test_tvdb_series_id__invalid_lang(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_series_id(
            tvdb_token,
            LOST_TVDB_ID_SERIES,
            language=Language(JUNK_TEXT, JUNK_TEXT, JUNK_TEXT),
            cache=False,
        )


def test_tvdb_series_id__invalid_id_imdb(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_series_id(tvdb_token, JUNK_TEXT, cache=False)


def test_tvdb_series_id__no_hits(tvdb_token):
    with pytest.raises(MnamerNotFoundException):
        tvdb_series_id(tvdb_token, LOST_TVDB_ID_SERIES * 2, cache=False)


def test_tvdb_series_id__success(tvdb_token):
    expected_top_level_keys = {
        "added",
        "addedBy",
        "airsDayOfWeek",
        "airsTime",
        "aliases",
        "banner",
        "fanart",
        "firstAired",
        "genre",
        "id",
        "imdbId",
        "language",
        "lastUpdated",
        "network",
        "networkId",
        "overview",
        "poster",
        "rating",
        "runtime",
        "season",
        "seriesId",
        "seriesName",
        "siteRating",
        "siteRatingCount",
        "slug",
        "status",
        "zap2itId",
    }

    result = tvdb_series_id(tvdb_token, LOST_TVDB_ID_SERIES)
    assert isinstance(result, dict)
    assert "data" in result
    assert set(result["data"].keys()) == expected_top_level_keys
    assert str(result["data"]["id"]) == LOST_TVDB_ID_SERIES
    assert result["data"]["seriesName"] == "Lost"


def test_tvdb_series_id__language(tvdb_token):
    result = tvdb_series_id(tvdb_token, THE_WITCHER_ID_SERIES, RUSSIAN_LANG)
    assert result["data"]["seriesName"] == "Ведьмак"


@pytest.mark.xfail(strict=False)
def test_tvdb_series_id_episodes__invalid_token():
    with pytest.raises(MnamerException):
        tvdb_series_id_episodes(JUNK_TEXT, LOST_TVDB_ID_SERIES, cache=False)


def test_tvdb_series_id_episodes__invalid_lang(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_series_id_episodes(
            tvdb_token,
            LOST_TVDB_ID_SERIES,
            language=Language(JUNK_TEXT, JUNK_TEXT, JUNK_TEXT),
            cache=False,
        )


def test_tvdb_series_id_episodes__invalid_id_imdb(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_series_id_episodes(tvdb_token, JUNK_TEXT, cache=False)


def test_tvdb_series_id_episodes__no_hits(tvdb_token):
    with pytest.raises(MnamerNotFoundException):
        tvdb_series_id_episodes(tvdb_token, LOST_TVDB_ID_SERIES * 2, cache=False)


def test_tvdb_series_id_episodes__success(tvdb_token):
    result = tvdb_series_id_episodes(tvdb_token, LOST_TVDB_ID_SERIES)
    assert isinstance(result, dict)
    assert "data" in result
    entry = result["data"][0]
    assert set(entry.keys()) == EXPECTED_TOP_LEVEL_SHOW_KEYS
    assert str(entry["id"]) == LOST_TVDB_ID_EPISODE


def test_tvdb_series_id_episodes__language(tvdb_token):
    result = tvdb_series_id_episodes(
        tvdb_token, THE_WITCHER_ID_SERIES, language=RUSSIAN_LANG
    )
    assert result["data"][0]["episodeName"] == "Начало конца"


@pytest.mark.xfail(strict=False)
def test_tvdb_series_id_episodes_query__invalid_token():
    with pytest.raises(MnamerException):
        tvdb_series_id_episodes_query(JUNK_TEXT, LOST_TVDB_ID_SERIES, cache=False)


def test_tvdb_series_id_episodes_query__invalid_lang(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_series_id_episodes_query(
            tvdb_token,
            LOST_TVDB_ID_SERIES,
            language=Language(JUNK_TEXT, JUNK_TEXT, JUNK_TEXT),
            cache=False,
        )


def test_tvdb_series_id_episodes_query__invalid_id_tvdb(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_series_id_episodes_query(tvdb_token, JUNK_TEXT, cache=False)


def test_tvdb_series_id_episodes_query__page_valid(tvdb_token):
    tvdb_series_id_episodes_query(tvdb_token, LOST_TVDB_ID_SERIES, page=1)
    tvdb_series_id_episodes_query(tvdb_token, LOST_TVDB_ID_SERIES, page=1, season=1)
    tvdb_series_id_episodes_query(
        tvdb_token, LOST_TVDB_ID_SERIES, page=1, season=1, episode=1
    )
    with pytest.raises(MnamerNotFoundException):
        tvdb_series_id_episodes_query(
            tvdb_token, LOST_TVDB_ID_SERIES, page=11, cache=False
        )
    with pytest.raises(MnamerNotFoundException):
        tvdb_series_id_episodes_query(
            tvdb_token, LOST_TVDB_ID_SERIES, page=2, season=1, cache=False
        )
    with pytest.raises(MnamerNotFoundException):
        tvdb_series_id_episodes_query(
            tvdb_token,
            LOST_TVDB_ID_SERIES,
            page=2,
            season=1,
            episode=1,
            cache=False,
        )


def test_tvdb_series_id_episodes_query__success_id_tvdb(tvdb_token):
    result = tvdb_series_id_episodes_query(tvdb_token, LOST_TVDB_ID_SERIES)
    assert isinstance(result, dict)
    assert "data" in result
    data = result["data"]
    assert len(data) == 100
    actual_top_level_keys = set(data[0].keys())
    assert actual_top_level_keys == EXPECTED_TOP_LEVEL_SHOW_KEYS
    assert str(data[0]["id"]) == LOST_TVDB_ID_EPISODE


def test_tvdb_series_id_episodes_query__success_id_tvdb_season(tvdb_token):
    result = tvdb_series_id_episodes_query(tvdb_token, LOST_TVDB_ID_SERIES, season=1)
    assert isinstance(result, dict)
    assert "data" in result
    data = result["data"]
    actual_top_level_keys = set(data[0].keys())
    assert actual_top_level_keys == EXPECTED_TOP_LEVEL_SHOW_KEYS
    assert str(data[0]["id"]) == LOST_TVDB_ID_EPISODE
    assert result["links"]["prev"] is None
    assert result["links"]["next"] is None


def test_tvdb_series_id_episodes_query__success_id_tvdb_season_episode(
    tvdb_token,
):
    result = tvdb_series_id_episodes_query(
        tvdb_token, LOST_TVDB_ID_SERIES, season=1, episode=1
    )
    assert isinstance(result, dict)
    assert "data" in result
    data = result["data"]
    actual_top_level_keys = set(data[0].keys())
    assert actual_top_level_keys == EXPECTED_TOP_LEVEL_SHOW_KEYS
    assert str(data[0]["id"]) == LOST_TVDB_ID_EPISODE
    assert result["links"]["prev"] is None
    assert result["links"]["next"] is None


def test_tvdb_series_id_episodes_query(tvdb_token):
    result = tvdb_series_id_episodes_query(
        tvdb_token,
        THE_WITCHER_ID_SERIES,
        season=1,
        episode=1,
        language=RUSSIAN_LANG,
    )
    assert result["data"][0]["episodeName"] == "Начало конца"


def test_tvdb_search_series__invalid_token():
    with pytest.raises(MnamerException):
        tvdb_search_series(JUNK_TEXT, "Lost", cache=False)


def test_tvdb_search_series__invalid_lang(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_search_series(
            tvdb_token,
            "Lost",
            language=Language(JUNK_TEXT, JUNK_TEXT, JUNK_TEXT),
            cache=False,
        )


def test_tvdb_search_series__invalid_id_imdb(tvdb_token):
    with pytest.raises(MnamerException):
        tvdb_search_series(tvdb_token, "Lost", id_imdb="xyz", cache=False)


def test_tvdb_search_series__success(tvdb_token):
    expected_top_level_keys = {
        "aliases",
        "banner",
        "firstAired",
        "id",
        "image",
        "network",
        "overview",
        "poster",
        "seriesName",
        "slug",
        "status",
    }
    result = tvdb_search_series(tvdb_token, "Lost")
    assert isinstance(result, dict)
    assert "data" in result
    data = result["data"]
    assert len(data) == 100
    actual_top_level_keys = set(data[0].keys())
    assert actual_top_level_keys == expected_top_level_keys


def test_tvdb_search_series__language(tvdb_token):
    results = tvdb_search_series(tvdb_token, "Witcher", language=RUSSIAN_LANG)
    assert any(result["seriesName"] for result in results["data"])



import pytest

from mnamer.endpoints import (
    tvmaze_episode_by_number,
    tvmaze_episodes_by_date,
    tvmaze_show,
    tvmaze_show_episodes_list,
    tvmaze_show_lookup,
    tvmaze_show_search,
    tvmaze_show_single_search,
)
from mnamer.exceptions import MnamerException, MnamerNotFoundException
from tests import EPISODE_META, JUNK_TEXT, TEST_DATE

pytestmark = [
    pytest.mark.network,
    pytest.mark.tvmaze,
    pytest.mark.flaky(reruns=1),
]

META = EPISODE_META["The Walking Dead"]
EXPECTED_SHOW_KEYS = [
    "_links",
    "dvdCountry",
    "externals",
    "genres",
    "id",
    "image",
    "language",
    "name",
    "network",
    "officialSite",
    "premiered",
    "rating",
    "runtime",
    "schedule",
    "status",
    "summary",
    "type",
    "updated",
    "url",
    "webChannel",
    "weight",
]

EXPECTED_EPISODE_KEYS = [
    "_links",
    "airdate",
    "airstamp",
    "airtime",
    "id",
    "image",
    "name",
    "number",
    "rating",
    "runtime",
    "season",
    "summary",
    "url",
]


def test_tvmaze_show():
    result = tvmaze_show(id_tvmaze=META["id_tvmaze"])
    assert result
    for expected_show_key in EXPECTED_SHOW_KEYS:
        assert expected_show_key in result
    assert result["id"] == META["id_tvmaze"]
    assert result["name"] == META["series"]


def test_tvmaze_show__embed_episodes():
    expected_show_keys = [
        "_embedded",
        "_links",
        "externals",
        "genres",
        "id",
        "image",
        "language",
        "name",
        "network",
        "officialSite",
        "premiered",
        "rating",
        "runtime",
        "schedule",
        "status",
        "summary",
        "type",
        "updated",
        "url",
        "webChannel",
        "weight",
    ]
    result = tvmaze_show(id_tvmaze=META["id_tvmaze"], embed_episodes=True)
    assert result
    for expected_show_key in expected_show_keys:
        assert expected_show_key in result
    assert result["id"] == META["id_tvmaze"]
    assert result["name"] == META["series"]


def test_tvmaze_show_search__success():
    results = tvmaze_show_search(META["series"])
    expected_top_level_keys = {"score", "show"}
    assert results
    result = results[0]
    assert expected_top_level_keys == set(result)
    for expected_show_key in EXPECTED_SHOW_KEYS:
        assert expected_show_key in result["show"]
    assert result["show"]["id"] == META["id_tvmaze"]
    assert result["show"]["name"] == META["series"]


def test_tvmaze_show_search__no_hits():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_show_search(JUNK_TEXT)


def test_tvmaze_show_single_search__success():
    result = tvmaze_show_single_search(META["series"])
    assert result
    for expected_show_key in EXPECTED_SHOW_KEYS:
        assert expected_show_key in result
    assert result["id"] == META["id_tvmaze"]
    assert result["name"] == META["series"]


def test_tvmaze_show_single_search__no_hits():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_show_single_search(JUNK_TEXT)


def test_tvmaze_show_lookup__idmb__success():
    result = tvmaze_show_lookup(id_imdb=META["id_imdb"])
    assert result
    for expected_show_key in EXPECTED_SHOW_KEYS:
        assert expected_show_key in result
    assert result["id"] == META["id_tvmaze"]
    assert result["name"] == META["series"]


def test_tvmaze_show_lookup__idmb__no_hits():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_show_lookup(id_imdb=JUNK_TEXT)


def test_tvmaze_show_lookup__tvdb__success():
    result = tvmaze_show_lookup(id_tvdb=META["id_tvdb"])
    assert result
    for expected_show_key in EXPECTED_SHOW_KEYS:
        assert expected_show_key in result
    assert result["id"] == META["id_tvmaze"]
    assert result["name"] == META["series"]


def test_tvmaze_show_lookup__tvdb__no_hits():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_show_lookup(id_tvdb=JUNK_TEXT)


def test_tvmaze_show_lookup__missing_id():
    with pytest.raises(MnamerException):
        tvmaze_show_lookup()


def test_tvmaze_show_lookup__both_ids():
    with pytest.raises(MnamerException):
        tvmaze_show_lookup(id_imdb=META["id_imdb"], id_tvdb=META["id_tvdb"])


def test_tvmaze_show_episodes_list__success():
    results = tvmaze_show_episodes_list(META["id_tvmaze"])
    assert results
    result = results[0]
    for expected_episode_key in EXPECTED_EPISODE_KEYS:
        assert expected_episode_key in result


def test_tvmaze_show_episodes_list__no_hits():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_show_episodes_list(JUNK_TEXT)


def test_tvmaze_episodes_by_date__success():
    results = tvmaze_episodes_by_date("247", TEST_DATE)
    assert results
    result = results[0]
    assert result["airdate"] == str(TEST_DATE)


def test_tvmaze_episodes_by_date__no_hits__bad_id():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_episodes_by_date(JUNK_TEXT, air_date=TEST_DATE)


def test_tvmaze_episodes_by_date__no_hits__bad_date():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_episodes_by_date(META["id_tvmaze"], TEST_DATE)


def test_tvmaze_episode_by_number__success():
    target_id = EPISODE_META["The Walking Dead"]["id_tvmaze"]
    result = tvmaze_episode_by_number(target_id, 3, 3)
    assert result
    assert result["id"] == 4116


def test_tvmaze_episode_by_number__no_hits():
    with pytest.raises(MnamerNotFoundException):
        tvmaze_episode_by_number(JUNK_TEXT, 1, 1)



"""A collection of utility functions non-specific to mnamer's domain logic."""

import datetime as dt
import json
import re
from os import walk
from os.path import exists, expanduser, expandvars, getsize, splitdrive, splitext
from pathlib import Path, PurePath
from typing import Any, Callable, Iterator
from unicodedata import normalize

import requests_cache
from requests.adapters import HTTPAdapter

from mnamer.const import CACHE_PATH, CURRENT_YEAR, SUBTITLE_CONTAINERS


def clean_dict(target_dict: dict, whitelist=None) -> dict:
    """Convenience function that removes a dicts keys that have falsy values."""
    return {
        str(k).strip(): str(v).strip()
        for k, v in target_dict.items()
        if v not in (None, Ellipsis, [], (), "") and (not whitelist or k in whitelist)
    }


def clear_cache():
    """Clears requests-cache cache."""
    get_session().cache.clear()


def crawl_in(file_paths: list[Path], recurse: bool = False) -> list[Path]:
    """Looks for files amongst or within paths provided."""
    found_files = set()
    for file_path in file_paths:
        if not file_path.exists():
            continue
        if file_path.is_file():
            found_files.add(Path(file_path).absolute())
            continue
        for root, _dirs, files in walk(file_path):
            for file in files:
                found_files.add(Path(root, file).absolute())
            if not recurse:
                break
    return sorted(list(found_files))


def crawl_out(filename: str | Path | PurePath) -> Path | None:
    """Looks for a file in the home directory and each directory up from cwd."""
    working_dir = Path.cwd()
    while True:
        parent_dir = working_dir.parent
        if parent_dir == working_dir:  # e.g. fs root or error
            break
        target = working_dir / filename
        if target.exists():
            return target
        working_dir = parent_dir
    target = Path.home() / filename
    return target if target.exists() else None


def filename_replace(filename: str, replacements: dict[str, str]) -> str:
    """Replaces keys in replacements dict with their values."""
    base, container = splitext(filename)
    base = str_replace(base, replacements)
    return base + container


def filter_blacklist(paths: list[Path], blacklist: list[str]) -> list[Path]:
    """Filters (set difference) paths by a collection of regex pattens."""
    return [
        path.absolute()
        for path in paths
        if not any(
            re.search(pattern, str(path), re.IGNORECASE)
            for pattern in blacklist
            if pattern
        )
    ]


def filter_containers(
    file_paths: list[Path], valid_containers: list[str]
) -> list[Path]:
    """Filters (set intersection) a collection of containers."""
    valid_containers = normalize_containers(valid_containers)
    return [
        file_path
        for file_path in file_paths
        if not valid_containers or file_path.suffix.lower() in valid_containers
    ]


def findall(s, ss) -> Iterator[int]:
    """Yields indexes of all start positions of substring matches in string."""
    i = s.find(ss)
    while i != -1:
        yield i
        i = s.find(ss, i + 1)


def fn_chain(*fn_list: Callable) -> Callable:
    """Chains a list of function calls into one."""
    return lambda *args, **kwargs: tuple(fn(*args, **kwargs) for fn in fn_list)


def fn_pipe(*fn_list: Callable) -> Callable:
    """Pipes a list of function calls into one."""

    def resolver(x):
        for fn in fn_list:
            x = fn(x)
        return x

    return resolver


def format_dict(body: dict) -> str:
    """
    Formats a dictionary into a multi-line bulleted string of key-value pairs.
    """
    return "\n".join([f" - {k} = {getattr(v, 'value', v)}" for k, v in body.items()])


def format_exception(body: Exception) -> str:
    return str(body)


def format_iter(body: list) -> str:
    """
    Formats an iterable into a multi-line bulleted string of its values.
    """
    return "\n".join(sorted([f" - {getattr(v, 'value', v)}" for v in body]))


def is_subtitle(container: str | Path | None) -> bool:
    """Returns True if container is a subtitle container."""
    if not container:
        return False
    return str(container).endswith(tuple(SUBTITLE_CONTAINERS))


def get_session() -> requests_cache.CachedSession:
    """Convenience function that returns request-cache session singleton."""

    def make_session():
        session = requests_cache.CachedSession(
            cache_name=str(CACHE_PATH),
            extension=".sqlite",
            expire_after=518_400,  # 6 days
        )
        adapter = HTTPAdapter(max_retries=3)
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        return session

    if hasattr(get_session, "session"):
        session: requests_cache.CachedSession = getattr(get_session, "session")
        return session
    session = make_session()
    setattr(get_session, "session", session)
    return session


def get_filesize(path: Path) -> str:
    """Returns the human-readable filesize for a given path."""
    size = float(getsize(path))
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024.0:
            break
        size /= 1024.0
    else:
        return "undetermined size"
    return f"{size:.{2}f}{unit}"


def json_dumps(d: dict[str, Any]) -> str:
    """A wrapper for json.dumps."""
    return json.dumps(
        {k: getattr(v, "value", v) for k, v in d.items()},
        allow_nan=False,
        check_circular=True,
        ensure_ascii=True,
        indent=4,
        skipkeys=True,
        sort_keys=True,
    )


def json_loads(path: str) -> dict[str, Any]:
    json_data = ""
    path = expanduser(path)
    path = expandvars(path)
    if exists(path):
        with open(path, mode="r") as fp:
            json_data = fp.read()
    return json.loads(json_data) if json_data else {}


def normalize_container(container: str) -> str:
    """Ensures all containers begin with a dot."""
    assert container
    if container and container[0] != ".":
        container = f".{container}"
    return container.lower()


def normalize_containers(container_list: list[str]) -> list[str]:
    """For a list of containers ensures that all containers begin with a dot."""
    return [normalize_container(container) for container in container_list]


def parse_date(value: str | dt.date | dt.datetime) -> dt.date:
    """Converts an ambiguously formatted date type into a date object."""
    if isinstance(value, str):
        value = value.replace("/", "-")
        value = value.replace(".", "-")
        value = dt.datetime.strptime(value, "%Y-%m-%d")
    if isinstance(value, dt.datetime):
        value = value.date()
    return value


def request_json(
    url,
    parameters: dict | list | None = None,
    body: dict | None = None,
    headers: dict | None = None,
    cache: bool = True,
) -> tuple[int, dict]:
    """
    Queries a url for json data.

    Note: Requests are cached using requests_cached for a week, this is done
    transparently by using the package's monkey patching.
    """
    assert url
    session = get_session()

    if isinstance(headers, dict):
        headers = clean_dict(headers)
    else:
        headers = dict()
    if isinstance(parameters, dict):
        parameters = [(k, v) for k, v in clean_dict(parameters).items()]
    if body:
        method = "POST"
        headers["content-type"] = "application/json"
        headers["content-length"] = str(len(body))
    else:
        method = "GET"
    headers["user-agent"] = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, "
        "like Gecko) Chrome/79.0.3945.88 Safari/537.36"
    )

    initial_cache_state = session._disabled  # yes, i'm a bad person
    try:
        session._disabled = not cache
        response = session.request(
            url=url,
            params=parameters,
            json=body,
            headers=headers,
            method=method,
            timeout=1,
        )
        status = response.status_code
        content = response.json() if status // 100 == 2 else None
    except:
        content = None
        status = 500
    finally:
        session._disabled = initial_cache_state
    return status, (content or {})


def str_fix_padding(s: str) -> str:
    """Truncates and collapses whitespace and delimiters in strings."""
    len_before = len(s)
    # Remove empty brackets
    s = re.sub(r"\(\s*\)", "", s)
    s = re.sub(r"\[\s*]", "", s)
    # Collapse dashes
    s = re.sub(r"-+", "-", s)
    # Collapse whitespace
    s = re.sub(r"\s+", " ", s)
    # Collapse repeating delimiters
    s = re.sub(r"( [-.,_])+", r"\1", s)
    # Strip leading/ trailing whitespace
    s = s.strip()
    # Strip leading/ trailing dashes
    s = s.strip("-")
    len_after = len(s)
    return s if len_before == len_after else str_fix_padding(s)


def str_replace(s: str, replacements: dict[str, str]) -> str:
    """Replaces keys in replacements dict with their values."""
    for word, replacement in replacements.items():
        if word in s:
            s = re.sub(re.escape(word), replacement, s, flags=re.IGNORECASE)
    return s


def str_replace_slashes(s: str) -> str:
    return str_replace(s, {"/": "-", "\\": "-"})


def str_sanitize(filename: str) -> str:
    """Removes illegal filename characters and condenses whitespace."""
    base, container = splitext(filename)
    if is_subtitle(container):
        base = base.rstrip(".")
        base, container_prefix = splitext(base)
        container = container_prefix + container
    base = re.sub(r"\s+", " ", base)
    drive, tail = splitdrive(base)
    tail = re.sub(r'[<>:"|?*&%=+@#`^]', "", tail)
    return drive + tail.strip("-., ") + container


def str_scenify(filename: str) -> str:
    """Replaces non ascii-alphanumerics with dots."""
    filename = normalize("NFKD", filename)
    filename.encode("ascii", "ignore")
    filename = re.sub(r"\s+", ".", filename)
    filename = re.sub(r"[^.\d\w/]", "", filename)
    filename = re.sub(r"\.+", ".", filename)
    return filename.lower().strip(".")


def str_title_case(s: str) -> str:
    """Attempts to intelligently apply title case transformations to strings."""

    if not s:
        return s

    lowercase_exceptions = {
        "a",
        "an",
        "and",
        "as",
        "at",
        "but",
        "by",
        "de",
        "des",
        "du",
        "for",
        "from",
        "in",
        "is",
        "le",
        "nor",
        "of",
        "on",
        "or",
        "the",
        "to",
        "un",
        "une",
        "with",
        "via",
    }
    uppercase_exceptions = {
        "i",
        "ii",
        "iii",
        "iv",
        "v",
        "vi",
        "vii",
        "viii",
        "ix",
        "x",
        "2d",
        "3d",
        "au",
        "aka",
        "atm",
        "bbc",
        "bff",
        "cia",
        "csi",
        "dc",
        "doa",
        "espn",
        "fbi",
        "ira",
        "jfk",
        "lol",
        "mlb",
        "mlk",
        "mtv",
        "nba",
        "nfl",
        "nhl",
        "nsfw",
        "nyc",
        "omg",
        "pga",
        "oj",
        "rsvp",
        "tnt",
        "tv",
        "ufc",
        "ufo",
        "uk",
        "usa",
        "vip",
        "wtf",
        "wwe",
        "wwi",
        "wwii",
        "xxx",
        "yolo",
    }
    padding_chars = ".- "
    paren_chars = "[](){}<>{}"
    punctuation_chars = paren_chars + "\"!?$,-.:;@_`'"
    partition_chars: str = padding_chars + punctuation_chars
    string_lower = s.lower()
    string_length = len(s)

    # uppercase first character
    s = s.lower()
    s = s[0].upper() + s[1:]

    # uppercase characters after padding characters
    for char in padding_chars:
        for pos in findall(s, char):
            if pos + 1 == string_length:
                break
            elif pos + 2 < string_length:
                s = s[: pos + 1] + s[pos + 1].upper() + s[pos + 2 :]
            else:
                s = s[: pos + 1] + s[pos + 1].upper()

    # uppercase characters inside parentheses
    for char in paren_chars:
        for pos in findall(s, char):
            if pos > 0 and s[pos - 1] not in padding_chars:
                continue
            elif pos + 1 < string_length:
                s = s[: pos + 1] + s[pos + 1].upper() + s[pos + 2 :]

    # process lowercase transformations
    for exception in lowercase_exceptions:
        for pos in findall(string_lower, exception):
            is_start = pos < 2
            if is_start:
                break
            prev_char = string_lower[pos - 1]
            is_left_partitioned = prev_char in padding_chars
            word_length = len(exception)
            ends = pos + word_length == string_length
            next_char = "" if ends else string_lower[pos + word_length]
            is_right_partitioned = not ends and next_char in padding_chars
            if is_left_partitioned and is_right_partitioned:
                s = s[:pos] + exception.lower() + s[pos + word_length :]

    # process uppercase transformations
    for exception in uppercase_exceptions:
        for pos in findall(string_lower, exception):
            is_start = pos == 0
            prev_char = None if is_start else string_lower[pos - 1]  # type: ignore
            is_left_partitioned = is_start or prev_char in partition_chars  # type: ignore
            word_length = len(exception)
            ends = pos + word_length == string_length
            next_char = "" if ends else string_lower[pos + word_length]
            is_right_partitioned = ends or next_char in partition_chars
            if is_left_partitioned and is_right_partitioned:
                s = s[:pos] + exception.upper() + s[pos + word_length :]

    return s


def year_parse(s: str) -> int | None:
    """Parses a year from a string."""
    regex = r"((?:19|20)\d{2})(?:$|[-/]\d{2}[-/]\d{2})"
    try:
        return int(re.findall(regex, str(s))[0])
    except IndexError:
        return None


def year_range_parse(years: str | int | None, tolerance: int = 1) -> tuple[int, int]:
    """Parses a year or dash-delimited year range."""
    regex = r"^((?:19|20)\d{2})?([-,: ]*)?((?:19|20)\d{2})?$"
    default_start = 1900
    default_end = CURRENT_YEAR
    try:
        start, dash, end = re.match(regex, str(years).strip()).groups()  # type: ignore
    except AttributeError:
        start, end, dash = None, None, True
    if not start and not end:
        start, end, dash = None, None, True
    start = int(start or default_start)
    end = int(end or default_end)
    if not dash:
        end = start
    return start - tolerance, end + tolerance



import pytest

from mnamer.argument import ArgLoader
from tests import JUNK_TEXT

pytestmark = [
    pytest.mark.e2e,
    pytest.mark.flaky(reruns=1),
]


def test_invalid_arguments(e2e_run):
    result = e2e_run(f"--{JUNK_TEXT}")
    assert result.code == 2
    assert result.out == f"invalid arguments: --{JUNK_TEXT}"


def test_no_arguments(e2e_run):
    result = e2e_run()
    assert result.code == 2
    assert result.out == ArgLoader().usage


@pytest.mark.usefixtures("setup_test_dir")
def test_no_matches_found(e2e_run, setup_test_files):
    setup_test_files("made up movie.mp4")
    result = e2e_run("-b", ".")
    assert result.code == 0
    assert "no matches found" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_no_files_found(e2e_run, setup_test_files):
    setup_test_files(
        "Avengers Infinity War/Avengers.Infinity.War.wmv",
        "Planet Earth II S01E06 - Cities (2016) (2160p).mp4",
        "game.of.thrones.01x05-eztv.mp4",
        "scan001.tiff",
    )
    result = e2e_run(JUNK_TEXT)
    assert result.code == 0
    assert "no media files found" in result.out



import pytest

from mnamer.exceptions import MnamerNotFoundException
from mnamer.metadata import MetadataMovie
from mnamer.providers import Omdb
from tests import JUNK_TEXT, MOVIE_META

pytestmark = [
    pytest.mark.network,
    pytest.mark.omdb,
    pytest.mark.flaky(reruns=1),
]


@pytest.fixture(scope="session")
def provider():
    return Omdb()


@pytest.mark.parametrize("meta", MOVIE_META.values(), ids=list(MOVIE_META))
def test_search__id(meta: dict[str, str], provider: Omdb):
    query = MetadataMovie(id_imdb=meta["id_imdb"])
    results = list(provider.search(query))
    assert len(results) == 1
    result = results[0]
    assert result.name == meta["name"]


def test_search__id__no_hits(provider: Omdb):
    query = MetadataMovie(id_imdb=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", MOVIE_META.values(), ids=list(MOVIE_META))
def test_search__name(meta, provider: Omdb):
    provider = Omdb()
    query = MetadataMovie(name=meta["name"])
    assert any(result.id_imdb == meta["id_imdb"] for result in provider.search(query))


@pytest.mark.parametrize("meta", MOVIE_META.values(), ids=list(MOVIE_META))
def test_search__name__year(meta, provider: Omdb):
    provider = Omdb()
    query = MetadataMovie(name=meta["name"], year=meta["year"])
    for result in provider.search(query):
        assert result.year
        assert (int(result.year) - int(meta["year"])) <= 5


def test_search__no_hits(provider: Omdb):
    query = MetadataMovie(name=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


def test_search__missing(provider: Omdb):
    query = MetadataMovie()
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))



from pathlib import Path
from unittest.mock import patch

import pytest
from requests import Session

from mnamer.const import CURRENT_YEAR, SUBTITLE_CONTAINERS
from mnamer.types import MediaType
from mnamer.utils import (
    clean_dict,
    crawl_in,
    crawl_out,
    filter_blacklist,
    filter_containers,
    fn_chain,
    fn_pipe,
    format_dict,
    format_exception,
    format_iter,
    is_subtitle,
    normalize_container,
    parse_date,
    request_json,
    str_fix_padding,
    str_replace,
    str_sanitize,
    str_scenify,
    str_title_case,
    year_parse,
    year_range_parse,
)
from tests import JUNK_TEXT, MockRequestResponse

pytestmark = pytest.mark.local

TEST_FILES: dict[str, Path] = {
    test_file: Path(*test_file.split("/"))
    for test_file in (
        "Avengers Infinity War/Avengers.Infinity.War.srt",
        "Avengers Infinity War/Avengers.Infinity.War.wmv",
        "Downloads/Return of the Jedi 1080p.mkv",
        "Downloads/archer.2009.s10e07.webrip.x264-lucidtv.mkv",
        "Downloads/the.goonies.1985.mp4",
        "Images/Photos/DCM0001.jpg",
        "Images/Photos/DCM0002.jpg",
        "Ninja Turtles (1990).mkv",
        "O.J. - Made in America S01EP03 (2016) (1080p).mp4",
        "Planet Earth II S01E06 - Cities (2016) (2160p).mp4",
        "Pride & Prejudice 2005.ts",
        "Quien a hierro mata [MicroHD 1080p][DTS 5.1-Castellano-AC3 5.1-Castellano+Subs][ES-EN]/Quien a hierro mata M1080.www.pctnew.org.mkv",
        "Sample/the mandalorian s01x02.mp4",
        "Skiing Trip.mp4",
        "aladdin.1992.avi",
        "aladdin.2019.avi",
        "archer.2009.s10e07.webrip.x264-lucidtv.mp4",
        "game.of.thrones.01x05-eztv.mp4",
        "homework.txt",
        "kill.bill.2003.ts",
        "lost s01e01-02.mp4",
        "made up movie.mp4",
        "made up show s01e10.mkv",
        "s.w.a.t.2017.s02e01.mkv",
        "scan001.tiff",
        "temp.zip",
    )
}
FILTER_FILENAMES = [path.absolute() for path in TEST_FILES.values()]
FILENAME_REPLACEMENT = "The quick brown fox jumps over the lazy dog"


def paths_for(*filenames: str):
    return [path.absolute() for name, path in TEST_FILES.items() if name in filenames]


def paths_except_for(*filenames: str):
    return [
        path.absolute() for name, path in TEST_FILES.items() if name not in filenames
    ]


# ------------------------------------------------------------------------------


def test_clean__dict_str_values():
    dict_in = {"apple": "pie", "candy": "corn", "bologna": "sandwich"}
    expected = dict_in
    actual = clean_dict(dict_in)
    assert actual == expected


def test_clean__dict_some_none():
    dict_in = {
        "super": "mario",
        "sonic": "hedgehog",
        "samus": None,
        "princess": "zelda",
        "bowser": None,
    }
    expected = {"super": "mario", "sonic": "hedgehog", "princess": "zelda"}
    actual = clean_dict(dict_in)
    assert actual == expected


def test_clean__dict_all_falsy():
    dict_in = {"who": None, "let": 0, "the": False, "dogs": [], "out": ()}
    expected = {"let": "0", "the": "False"}
    actual = clean_dict(dict_in)
    assert actual == expected


def test_clean__dict_int_values():
    dict_in = {"0": 0, "1": 1, "2": 2, "3": 3, "4": 4}
    expected = {"0": "0", "1": "1", "2": "2", "3": "3", "4": "4"}
    actual = clean_dict(dict_in)
    assert actual == expected


def test_clean__dict_str_strip():
    dict_in = {
        "please": ".",
        "fix ": ".",
        " my spacing": ".",
        "  issues  ": ".",
    }
    expected = {"please": ".", "fix": ".", "my spacing": ".", "issues": "."}
    actual = clean_dict(dict_in)
    assert actual == expected


def test_clean__dict_whitelist():
    whitelist = {"apple", "raspberry", "pecan"}
    dict_in = {"apple": "pie", "pecan": "pie", "pumpkin": "pie"}
    expected = {"apple": "pie", "pecan": "pie"}
    actual = clean_dict(dict_in, whitelist)
    assert actual == expected


def test_parse_date__slash():
    result = parse_date("2000/10/30")
    assert result.year == 2000
    assert result.month == 10
    assert result.day == 30


def test_parse_date__dot():
    result = parse_date("2000.10.30")
    assert result.year == 2000
    assert result.month == 10
    assert result.day == 30


def test_parse_date__dash():
    result = parse_date("2000-10-30")
    assert result.year == 2000
    assert result.month == 10
    assert result.day == 30


@pytest.mark.usefixtures("setup_test_dir")
def test_dir_crawl_in__files__none():
    expected = []
    actual = crawl_in([Path(".", JUNK_TEXT)])
    assert actual == expected


@pytest.mark.usefixtures("setup_test_dir")
def test_dir_crawl_in__files__flat(setup_test_files):
    setup_test_files(*TEST_FILES.keys())
    expected = paths_for(
        "Ninja Turtles (1990).mkv",
        "O.J. - Made in America S01EP03 (2016) (1080p).mp4",
        "Planet Earth II S01E06 - Cities (2016) (2160p).mp4",
        "Pride & Prejudice 2005.ts",
        "Skiing Trip.mp4",
        "aladdin.1992.avi",
        "aladdin.2019.avi",
        "archer.2009.s10e07.webrip.x264-lucidtv.mp4",
        "avengers infinity war.wmv",
        "game.of.thrones.01x05-eztv.mp4",
        "homework.txt",
        "kill.bill.2003.ts",
        "lost s01e01-02.mp4",
        "made up movie.mp4",
        "made up show s01e10.mkv",
        "s.w.a.t.2017.s02e01.mkv",
        "scan001.tiff",
        "temp.zip",
    )
    actual = crawl_in([Path.cwd()])
    assert set(actual) == set(expected)


@pytest.mark.usefixtures("setup_test_dir")
def test_dir_crawl_in__dirs__multiple(setup_test_files):
    setup_test_files(*TEST_FILES.keys())
    file_paths = [Path(filename) for filename in ("Desktop", "Downloads", "Images")]
    actual = crawl_in(file_paths)
    expected = paths_for(
        "Downloads/Return of the Jedi 1080p.mkv",
        "Downloads/archer.2009.s10e07.webrip.x264-lucidtv.mkv",
        "Downloads/the.goonies.1985.mp4",
        "Downloads/the.goonies.1985.sample.mp4",
    )
    assert set(actual) == set(expected)


@pytest.mark.usefixtures("setup_test_dir")
def test_dir_crawl_in__dirs__recurse(setup_test_files):
    setup_test_files(*TEST_FILES.keys())
    actual = crawl_in([Path.cwd()], recurse=True)
    expected = paths_for(*TEST_FILES.keys())
    assert set(actual) == set(expected)


@pytest.mark.usefixtures("setup_test_dir")
def test_test_crawl_out__walking(setup_test_files):
    setup_test_files(*TEST_FILES.keys())
    expected = Path("aladdin.2019.avi").absolute()
    actual = crawl_out("aladdin.2019.avi")
    assert actual == expected


@pytest.mark.usefixtures("setup_test_dir")
def test_test_crawl_out__no_match():
    path = Path("/", "some_path", "avengers infinity war.wmv")
    assert crawl_out(path) is None


def test_str_replace__no_change():
    replacements = {}
    expected = FILENAME_REPLACEMENT
    actual = str_replace(FILENAME_REPLACEMENT, replacements)
    assert actual == expected


def test_str_replace__single_replacement():
    replacements = {"brown": "red"}
    expected = "The quick red fox jumps over the lazy dog"
    actual = str_replace(FILENAME_REPLACEMENT, replacements)
    assert actual == expected


def test_str_replace__multiple_replacement():
    replacements = {"the": "a", "lazy": "misunderstood", "dog": "cat"}
    expected = "a quick brown fox jumps over a misunderstood cat"
    actual = str_replace(FILENAME_REPLACEMENT, replacements)
    assert actual == expected


def test_str_replace__regex_escaping():
    expected = "hello, world!"
    actual = str_replace("hello, world?", {"?": "!"})
    assert actual == expected


def test_str_sanitize__condense_whitespace():
    filename = "fix  these    spaces\tplease "
    expected = "fix these spaces please"
    actual = str_sanitize(filename)
    assert actual == expected


def test_str_sanitize__remove_illegal_chars():
    filename = "sup*:>"
    expected = "sup"
    actual = str_sanitize(filename)
    assert actual == expected


@pytest.mark.parametrize("filename", ("xx.mkv.srt", "xx..mkv.srt", "xx.mkv..srt"))
def test_srt_sanitize__subtitle(filename):
    expected = "xx.mkv.srt"
    actual = str_sanitize(filename)
    assert actual == expected


def test_str_scenify__dot_concat():
    filename = "some  file..name"
    expected = "some.file.name"
    actual = str_scenify(filename)
    assert actual == expected


def test_str_scenify__remove_non_alpanum_chars():
    filename = "who let the dogs out!? (1999)"
    expected = "who.let.the.dogs.out.1999"
    actual = str_scenify(filename)
    assert actual == expected


def test_str_scenify__spaces_to_dots():
    filename = " Space Jam "
    expected = "space.jam"
    actual = str_scenify(filename)
    assert actual == expected


def test_str_scenify__utf8_to_ascii():
    filename = "Amélie"
    expected = "amelie"
    actual = str_scenify(filename)
    assert actual == expected


@pytest.mark.parametrize("sequence", (list(), set(), tuple()))
def test_filter_blacklist__filter_none(sequence):
    expected = FILTER_FILENAMES
    actual = filter_blacklist(FILTER_FILENAMES, sequence)
    assert actual == expected


def test_filter_blacklist__filter_multiple_paths_single_pattern():
    expected = paths_except_for(
        "Images/Photos/DCM0001.jpg", "Images/Photos/DCM0002.jpg"
    )
    actual = filter_blacklist(FILTER_FILENAMES, ["dcm"])
    assert actual == expected


def test_filter_blacklist__filter_multiple_paths_multiple_patterns():
    expected = paths_except_for(
        "temp.zip",
        "Downloads/the.goonies.1985.sample.mp4",
        "Sample/the mandalorian s01x02.mp4",
    )
    actual = filter_blacklist(FILTER_FILENAMES, ["temp", "sample"])
    assert actual == expected


def test_filter_blacklist__filter_single_path_single_pattern():
    expected = paths_except_for(
        "Images/sample.file.mp4", "Sample/the mandalorian s01x02.mp4"
    )
    actual = filter_blacklist(expected, ["sample"])
    assert actual == expected


def test_filter_blacklist__filter_single_path_multiple_patterns():
    expected = paths_except_for(
        "Images/sample.file.mp4", "Sample/the mandalorian s01x02.mp4"
    )
    actual = filter_blacklist(expected, ["files", "sample"])
    assert expected == actual


def test_filter_blacklist__regex():
    pattern = r"\s+"
    expected = paths_except_for(
        "Avengers Infinity War/Avengers.Infinity.War.srt",
        "Avengers Infinity War/Avengers.Infinity.War.wmv",
        "Downloads/Return of the Jedi 1080p.mkv",
        "Ninja Turtles (1990).mkv",
        "O.J. - Made in America S01EP03 (2016) (1080p).mp4",
        "Planet Earth II S01E06 - Cities (2016) (2160p).mp4",
        "Pride & Prejudice 2005.ts",
        "Quien a hierro mata [MicroHD 1080p][DTS 5.1-Castellano-AC3 5.1-Castellano+Subs][ES-EN]/Quien a hierro mata M1080.www.pctnew.org.mkv",
        "Sample/the mandalorian s01x02.mp4",
        "Skiing Trip.mp4",
        "lost s01e01-02.mp4",
        "made up movie.mp4",
        "made up show s01e10.mkv",
    )
    actual = filter_blacklist(FILTER_FILENAMES, [pattern])
    assert actual == expected


def test_filter_containers__filter_none():
    expected = FILTER_FILENAMES
    actual = filter_containers(FILTER_FILENAMES, [])
    assert expected == actual


@pytest.mark.parametrize("containers", (["jpg"], [".jpg"]))
def test_filter_containers__filter_multiple_paths_single_pattern(
    containers: list[str],
):
    expected = paths_for("Images/Photos/DCM0001.jpg", "Images/Photos/DCM0002.jpg")
    actual = filter_containers(FILTER_FILENAMES, containers)
    assert expected == actual


@pytest.mark.parametrize("containers", (["mkv", "zip"], [".mkv", ".zip"]))
def test_filter_containers__filter_multiple_paths_multi_pattern(
    containers: list[str],
):
    expected = paths_for(
        "Desktop/temp.zip",
        "Downloads/Return of the Jedi 1080p.mkv",
        "Downloads/archer.2009.s10e07.webrip.x264-lucidtv.mkv",
        "Ninja Turtles (1990).mkv",
        "Quien a hierro mata [MicroHD 1080p][DTS 5.1-Castellano-AC3 5.1-Castellano+Subs][ES-EN]/Quien a hierro mata M1080.www.pctnew.org.mkv",
        "made up show s01e10.mkv",
        "s.w.a.t.2017.s02e01.mkv",
        "temp.zip",
    )
    actual = filter_containers(FILTER_FILENAMES, containers)
    assert expected == actual


@pytest.mark.parametrize("containers", (["mp4", "zip"], [".mp4", ".zip"]))
def test_filter_containers__filter_single_path_multi_pattern(
    containers: list[str],
):
    filepaths = paths_for("Images/Skiing Trip.mp4")
    expected = filepaths
    actual = filter_containers(filepaths, containers)
    assert expected == actual


def test_fn_chain():
    def add(x, y):
        return x + y

    def multiply(x, y):
        return x * y

    def exponent(x, y):
        return x**y

    expected = (7 + 8, 7 * 8, 7**8)
    actual = fn_chain(add, multiply, exponent)(7, 8)
    assert actual == expected


def test_fn_pipe():
    def add(x):
        return x + 5

    def multiply(x):
        return x * 5

    def exponent(x):
        return x**5

    expected = ((8 + 5) * 5) ** 5
    actual = fn_pipe(add, multiply, exponent)(8)
    assert actual == expected


def test_format_dict():
    d = {1: "a", 2: "b", 3: "c"}
    expected = " - 1 = a\n" + " - 2 = b\n" + " - 3 = c"
    actual = format_dict(d)
    assert actual == expected


def test_format_dict__enum():
    d = {1: MediaType.EPISODE, 2: MediaType.MOVIE}
    expected = " - 1 = episode\n" + " - 2 = movie"
    actual = format_dict(d)
    assert actual == expected


def test_format_exception():
    e = Exception("hello, world!")
    expected = "hello, world!"
    actual = format_exception(e)
    assert actual == expected


@pytest.mark.parametrize("it", ((1, 3, 2), [1, 3, 2], {1, 3, 2}))
def test_format_iter(it):
    expected = " - 1\n" + " - 2\n" + " - 3"
    actual = format_iter(it)
    assert actual == expected


def test_format_iter__enum():
    it = [MediaType.MOVIE, MediaType.EPISODE]
    expected = " - episode\n" + " - movie"
    actual = format_iter(it)
    assert actual == expected


@pytest.mark.parametrize("container", SUBTITLE_CONTAINERS)
def test_is_subtitle__true(container):
    assert is_subtitle(container) is True


@pytest.mark.parametrize("container", (None, "", ".abc123", "srt"))
def test_is_subtitle__false(container):
    assert is_subtitle(container) is False


def test_normalize_container__has_no_dot():
    expected = ".mkv"
    actual = normalize_container("mkv")
    assert actual == expected


def test_normalize_container__has_dot():
    expected = ".mkv"
    actual = normalize_container(".mkv")
    assert actual == expected


@pytest.mark.parametrize("code", [200, 201, 209, 400, 500])
@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__status(mock_request, code):
    mock_response = MockRequestResponse(code, "{}")
    mock_request.return_value = mock_response
    status, _ = request_json("http://...", cache=False)
    assert status == code


@pytest.mark.parametrize(
    "code, truthy", [(200, True), (299, True), (400, False), (500, False)]
)
@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__data(mock_request, code, truthy):
    mock_response = MockRequestResponse(code, '{"status":true}')
    mock_request.return_value = mock_response
    _, content = request_json("http://...", cache=False)
    assert content if truthy else not content


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__json_data(mock_request):
    json_data = """{
        "status": true,
        "data": {
            "title": "The Matrix",
            "year": 1999,
            "genre": null
        }
    }"""
    json_dict = {
        "status": True,
        "data": {"title": "The Matrix", "year": 1999, "genre": None},
    }
    mock_response = MockRequestResponse(200, json_data)
    mock_request.return_value = mock_response
    status, content = request_json("http://...", cache=False)
    assert status == 200
    assert content == json_dict


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__xml_data(mock_request):
    xml_data = """
        <?xml version="1.0" encoding="UTF-8" ?>
        <status>true</status>
        <data>
            <title>The Matrix</title>
            <year>1999</year>
            <genre />
        </data>
    """

    mock_response = MockRequestResponse(200, xml_data)
    mock_request.return_value = mock_response
    status, content = request_json("http://...", cache=False)
    assert status == 500
    assert content == {}


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__html_data(mock_request):
    html_data = """
        <!DOCTYPE html>
        <html>
            <body>
                <h1>Data</h1>
                <ul>
                <li>Title: The Matrix</li>
                <li>Year: 1999</li>
                <li>Genre: ???</li>
                </ul>
            </body>
        </html>
    """
    mock_response = MockRequestResponse(200, html_data)
    mock_request.return_value = mock_response
    status, content = request_json("http://...", cache=False)
    assert status == 500
    assert content == {}


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__get_headers(mock_request):
    mock_request.side_effect = Session().request
    request_json(url="http://google.com", headers={"apple": "pie", "orange": None})
    _, kwargs = mock_request.call_args
    assert kwargs["method"] == "GET"
    assert len(kwargs["headers"]) == 2
    assert kwargs["headers"]["apple"] == "pie"
    assert "user-agent" in kwargs["headers"]


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__get_parameters(mock_request):
    test_parameters = {"apple": "pie"}
    mock_request.side_effect = Session().request
    request_json(url="http://google.com", parameters=test_parameters)
    _, kwargs = mock_request.call_args
    assert kwargs["method"] == "GET"
    assert kwargs["params"][0] == ("apple", "pie")


def test_request_json__get_invalid_url():
    status, content = request_json("mapi rulez", cache=False)
    assert status == 500
    assert content == {}


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__post_body(mock_request):
    data = {"apple": "pie"}
    mock_request.side_effect = Session().request
    request_json(url="http://google.com", body=data)
    _, kwargs = mock_request.call_args
    assert kwargs["method"] == "POST"
    assert data == kwargs["json"]


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__post_parameters(mock_request):
    mock_request.side_effect = Session().request
    data = {"apple": "pie", "orange": None}
    request_json(url="http://google.com", body=data, parameters=data)
    _, kwargs = mock_request.call_args
    assert kwargs["method"] == "POST"
    assert kwargs["params"][0] == ("apple", "pie")


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__post_headers(mock_request):
    mock_request.side_effect = Session().request
    data = {"apple": "pie", "orange": None}
    request_json(url="http://google.com", body=data, headers=data)
    _, kwargs = mock_request.call_args
    assert kwargs["method"] == "POST"
    assert "apple" in kwargs["headers"]
    assert "orange" not in kwargs["headers"]


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_request_json__failure(mock_request):
    mock_request.side_effect = Exception
    status, content = request_json(url="http://google.com")
    assert status == 500
    assert content == {}


@pytest.mark.parametrize("s", ("()x", "x()", "()[]x", "[]x()()"))
def test_str_fix_padding__strip_empty_brackets(s: str):
    expected = "x"
    actual = str_fix_padding(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("-y", "y-", "--y", "-----y----"))
def test_str_fix_padding__collapse_dashes(s: str):
    expected = "y"
    actual = str_fix_padding(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("a  b", "   a b", "a\t \tb ", "    a    b "))
def test_str_fix_padding__collapse_whitespace(s: str):
    expected = "a b"
    actual = str_fix_padding(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("s - t", "s   -- t", "s - t", "s - - - t"))
def test_str_fix_padding__collapse_delimiters(s: str):
    expected = "s - t"
    actual = str_fix_padding(s)
    assert actual == expected


def test_str_fix_padding__empty():
    expected = ""
    actual = str_fix_padding("")
    assert actual == expected


@pytest.mark.parametrize("s", ("jack and jill", "JACK AND JILL"))
def test_str_title_case__lower(s: str):
    expected = "Jack and Jill"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("a wrinkle in time", "A WRINKLE IN TIME"))
def test_str_title_case__lower__starts_with(s: str):
    expected = "A Wrinkle in Time"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("where to", "WHERE TO"))
def test_str_title_case__lower__ends_with(s: str):
    expected = "Where To"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("at the theatre", "AT THE THEATRE"))
def test_str_title_case__lower__only_whole_words(s: str):
    expected = "At the Theatre"  # theatre prefixed with 'the'
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("world war ii", "WORLD WAR II"))
def test_str_title_case__upper(s: str):
    expected = "World War II"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("ixx", "IXX"))
def test_str_title_case__upper_only_whole_worlds(s: str):
    expected = "Ixx"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("ufo sighting", "UFO SIGHTING"))
def test_str_title_case__upper__starts_with(s: str):
    expected = "UFO Sighting"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("sighted ufo", "SIGHTED UFO"))
def test_str_title_case__upper__ends_with(s: str):
    expected = "Sighted UFO"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("i ii iii iv", "I II III IV"))
def test_str_title_case__upper_multiple(s: str):
    expected = "I II III IV"
    actual = str_title_case(s)
    assert actual == expected


def test_str_title_case__letter_before_punctuation():
    expected = "Kill Bill Vol. 2"
    actual = str_title_case("kill bill vol. 2")
    assert actual == expected


@pytest.mark.parametrize("s", ("spider-man", "SPIDER-MAN"))
def test_str_title__after_punctuation__middle(s: str):
    expected = "Spider-Man"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize("s", ("spider-", "SPIDER-"))
def test_str_title__after_punctuation__end(s: str):
    expected = "Spider-"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize(
    "s",
    ("Don't stop", "DON'T STOP", "don't stop"),
)
def test_str_title_case__apostrophes(s):
    expected = "Don't Stop"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize(
    "s",
    (
        "Who Let the Dog (or Dogs) out",
        "WHO LET THE DOG (OR DOGS) OUT",
        "who let the dog (or dogs) out",
    ),
)
def test_str_title__padding__space(s):
    expected = "Who Let the Dog (Or Dogs) Out"
    actual = str_title_case(s)
    assert actual == expected


@pytest.mark.parametrize(
    "s",
    ("Who Let the Dog(s) out", "WHO LET THE DOG(S) OUT", "who let the dog(s) out"),
)
def test_str_title__padding__no_space(s):
    expected = "Who Let the Dog(s) Out"
    actual = str_title_case(s)
    assert actual == expected


def test_str_title_case__empty():
    expected = ""
    actual = str_title_case("")
    assert actual == expected


def test_str_title_case__single_char():
    expected = "A"
    actual = str_title_case("a")
    assert actual == expected


def test_year_parse__valid():
    expected = 1987
    actual = year_parse("1987")
    assert actual == expected


@pytest.mark.parametrize("s", ("1", "5000", "", " hello", "-", ","))
def test_year_parse__unexpected(s: str):
    assert year_parse(s) is None


@pytest.mark.parametrize("s", ("1950", " 1950", "  1950 "))
@pytest.mark.parametrize("t", (0, 10, 100))
def test_year_range_parse__exact(t: int, s: str):
    expected = (1950 - t, 1950 + t)
    actual = year_range_parse(s, t)
    assert actual == expected


@pytest.mark.parametrize("s", ("1990-2000", "1990 -  2000", "1990,2000", "1990:2000"))
@pytest.mark.parametrize("t", (0, 10, 100))
def test_year_range_parse__has_start_has_end(t: int, s: str):
    expected = (1990 - t, 2000 + t)
    actual = year_range_parse(s, t)
    assert actual == expected


@pytest.mark.parametrize("s", ("1990-", "1990 - ", "1990:"))
@pytest.mark.parametrize("t", (0, 10, 100))
def test_year_range_parse__has_start_no_end(t: int, s: str):
    expected = (1990 - t, CURRENT_YEAR + t)
    actual = year_range_parse(s, t)
    assert actual == expected


@pytest.mark.parametrize("s", ("-2005", " : 2005", ",2005"))
@pytest.mark.parametrize("t", (0, 10, 100))
def test_year_range_parse__no_start_has_end(t: int, s: str):
    expected = (1900 - t, 2005 + t)
    actual = year_range_parse(s, t)
    assert actual == expected


@pytest.mark.parametrize("s", ("15-5000", "", " hello", "-", ","))
@pytest.mark.parametrize("t", (0, 10, 100))
def test_year_range_parse__unexpected(t: int, s: str):
    expected = (1900 - t, CURRENT_YEAR + t)
    actual = year_range_parse(s, t)
    assert actual == expected



import sys
from os import chdir
from pathlib import Path

import pytest
from teletype.io import strip_format

from mnamer.exceptions import MnamerException
from mnamer.frontends import Cli
from mnamer.setting_store import SettingStore
from mnamer.target import Target
from tests import E2EResult

# Move up to root directory if run from subdirectory
cwd = Path().resolve()
while not (cwd / "pyproject.toml").exists():
    assert cwd != cwd.parent, "could not determine testing root"
    cwd = cwd.parent
    chdir(cwd)

# Create E2E test log
E2E_LOG = Path().absolute() / "e2e.log"
E2E_LOG.open("w").close()


@pytest.fixture(autouse=True)
def reset_args():
    """Clears argv before and after running test."""
    del sys.argv[:]
    sys.argv.append("mnamer")


@pytest.fixture
def e2e_run(capsys, request):
    """Runs main with provided arguments and returns stdout."""

    def fn(*args):
        Target.reset_providers()
        out = ""
        code = 0
        for arg in args:
            sys.argv.append(arg)
        try:
            settings = SettingStore()
            settings.load()
            Cli(settings).launch()
        except MnamerException as e:
            out += str(e)
            code = 2
        except SystemExit as e:
            code = e.code
        out += strip_format(capsys.readouterr().out.strip())
        out += strip_format(capsys.readouterr().err.strip())
        with open(E2E_LOG, "a+") as fp:
            fp.write("=" * 10 + "\n")
            fp.write(request.node.name + "\n")
            fp.write("-" * 10 + "\n")
            fp.write(out + "\n\n")
        return E2EResult(code, out)

    return fn



import datetime as dt

import pytest

from mnamer.metadata import Metadata, MetadataEpisode, MetadataMovie

pytestmark = pytest.mark.local

TEXT_CASES = ["test", "Test", "TEST", "TeSt"]


@pytest.mark.parametrize("value", TEXT_CASES)
def test_metadata__convert_synopsis(value):
    metadata = Metadata(synopsis=value)
    assert metadata.synopsis == "Test"


@pytest.mark.parametrize("value", TEXT_CASES)
def test_metadata__convert_container(value):
    metadata = Metadata(container=value)
    assert metadata.container == ".test"


@pytest.mark.parametrize("value", TEXT_CASES)
def test_metadata__convert_group(value):
    metadata = Metadata(group=value)
    assert metadata.group == "TEST"


@pytest.mark.parametrize("value", TEXT_CASES)
def test_metadata__convert_quality(value):
    metadata = Metadata(quality=value)
    assert metadata.quality == "test"


def test_metadata__format():
    with pytest.raises(NotImplementedError):
        format(Metadata())


@pytest.mark.parametrize("value", ["1987/10/01", "1987-10-01", "1987.10.01"])
def test_metadata_episode__convert__date(value):
    metadata = MetadataEpisode(date=value)
    assert metadata.date == dt.date(1987, 10, 1)


@pytest.mark.parametrize("value", TEXT_CASES)
def test_metadata_episode__convert_series(value):
    metadata = MetadataEpisode(series=value)
    assert metadata.series == "Test"


def test_metadata_episode__convert_season():
    metadata = MetadataEpisode(season=1)
    assert metadata.season == 1


def test_metadata_episode__convert_episode():
    metadata = MetadataEpisode(episode=1)
    assert metadata.episode == 1


def test_metadata_episode__format_default():
    metadata = MetadataEpisode(series="Spongebob Squarepants", season=4, episode=4)
    actual = format(metadata)
    expected = "Spongebob Squarepants - 04x04"
    assert actual == expected


def test_metadata_episode__specials__00x00():
    metadata = MetadataEpisode(season=0, episode=1, series="Parks and Recreation")
    actual = format(metadata, "{series} - {season:02}x{episode:02}")
    expected = "Parks and Recreation - 00x01"
    assert actual == expected


def test_metadata_episode__specials__s00e00():
    metadata = MetadataEpisode(season=0, episode=1, series="Parks and Recreation")
    actual = format(metadata, "{series} - S{season:02}E{episode:02}")
    expected = "Parks and Recreation - S00E01"
    assert actual == expected


def test_metadata_episode__format_with_specifiers():
    metadata = MetadataEpisode(series="Spongebob Squarepants", season=4, episode=4)
    format_spec = "{series[0]}/{series} season={season} episode={episode:04}"
    actual = format(metadata, format_spec)
    expected = "S/Spongebob Squarepants season=4 episode=0004"
    assert actual == expected


@pytest.mark.parametrize("value", TEXT_CASES)
def test_metadata_movie__convert_name(value):
    metadata = MetadataMovie(name=value)
    assert metadata.name == "Test"


def test_metadata_movie__format_default():
    metadata = MetadataMovie(name="Napoleon Dynamite", year="2004")
    expected = "Napoleon Dynamite (2004)"
    actual = format(metadata)
    assert actual == expected


@pytest.mark.parametrize(
    "format_spec",
    (
        "{name} - {year} - {group}",
        "{group} - {name} - {year}",
        "{name} - {group} - {year}",
    ),
)
def test_metadata_movie__format_missing_field(format_spec: str):
    metadata = MetadataMovie(name="jojo rabbit", year="2019")
    expected = "Jojo Rabbit - 2019"
    actual = format(metadata, format_spec)
    assert actual == expected


def test_metadata_movie__format_with_specifiers():
    metadata = MetadataMovie(name="pineapple express")
    format_spec = "{name[0]}/{name}"
    expected = "P/Pineapple Express"
    actual = format(metadata, format_spec)
    assert actual == expected



from pathlib import Path

import pytest

from mnamer.const import SUBTITLE_CONTAINERS

pytestmark = [
    pytest.mark.e2e,
    pytest.mark.flaky(reruns=2, reruns_delay=5),
]


@pytest.mark.usefixtures("setup_test_dir")
def test_complex_metadata(e2e_run, setup_test_files):
    setup_test_files(
        "Quien a hierro mata [MicroHD 1080p][DTS 5.1-Castellano-AC3 5.1-Castellano+Subs][ES-EN].mkv"
    )
    result = e2e_run("--batch", "--media=movie", ".")
    assert result.code == 0
    assert "Quien a Hierro Mata (2019).mkv" in result.out
    assert "1 out of 1 files processed successfully" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_absolute_path(e2e_run, setup_test_files):
    setup_test_files("kill.bill.vol.1.mkv")
    result = e2e_run("--batch", "--media=movie", "kill.bill.vol.1.mkv")
    assert result.code == 0
    assert "Kill Bill Vol. 1 (2003).mkv" in result.out
    assert "1 out of 1 files processed successfully" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_batch(e2e_run, setup_test_files):
    setup_test_files(
        "Avengers Infinity War/Avengers.Infinity.War.srt",
        "Avengers Infinity War/Avengers.Infinity.War.wmv",
        "Downloads/the.goonies.1985.mp4",
        "Images/Photos/DCM0001.jpg",
        "aladdin.2019.avi",
        "archer.2009.s10e07.webrip.x264-lucidtv.mp4",
        "game.of.thrones.01x05-eztv.mp4",
        "lost s01e01-02.mp4",
    )
    result = e2e_run("--batch", ".")
    assert result.code == 0
    assert "4 out of 4 files processed successfully" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_lower(e2e_run, setup_test_files):
    setup_test_files("aladdin.2019.avi")
    result = e2e_run("--batch", "--lower", ".")
    assert result.code == 0
    assert "aladdin (2019).avi" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_scene(e2e_run, setup_test_files):
    setup_test_files("aladdin.2019.avi")
    result = e2e_run("--batch", "--scene", ".")
    assert result.code == 0
    assert "aladdin.2019.avi" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_no_guess(e2e_run, setup_test_files):
    setup_test_files(
        "game.of.thrones.01x05-eztv.mp4",
        "made up movie.mp4",
        "made up show s01e10.mkv",
    )
    result = e2e_run("--noguess", "--batch", ".")
    assert result.code == 0
    assert result.out.count("skipping (--no-guess)") == 2


@pytest.mark.usefixtures("setup_test_dir")
def test_no_overwrite(e2e_run, setup_test_files):
    setup_test_files("aladdin.1992.avi", "Aladdin-1992.avi")
    result = e2e_run("--no-overwrite", "--batch", ".")
    assert result.code == 0
    assert result.out.count("skipping (--no-overwrite)") == 1
    assert "1 out of 2 files processed successfully" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_ignore(e2e_run, setup_test_files):
    setup_test_files(
        "Downloads/the.goonies.1985.mkv",
        "O.J. - Made in America S01EP03 (2016) (1080p).mp4",
        "aladdin.2019.avi",
        "kill.bill.2003.ts",
    )
    result = e2e_run("--batch", r"--ignore=.+(?:mkv|mp4)", ".")
    assert result.code == 0
    assert ".mkv" not in result.out
    assert ".mp4" not in result.out
    assert "2 out of 2 files processed successfully" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_mask(e2e_run, setup_test_files):
    setup_test_files(
        "Downloads/the.goonies.1985.mkv",
        "O.J. - Made in America S01EP03 (2016) (1080p).mp4",
        "aladdin.2019.avi",
        "kill.bill.2003.ts",
    )
    result = e2e_run("--batch", "--mask=mp4", ".")
    assert result.code == 0
    assert ".mkv" not in result.out
    assert ".avi" not in result.out
    assert ".ts" not in result.out
    assert ".mp4" in result.out
    assert "1 out of 1 files processed successfully" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_multi_part_episode(e2e_run, setup_test_files):
    setup_test_files("lost s01e01-02.mp4")
    result = e2e_run("--batch", ".")
    assert result.code == 0
    assert ".mkv" not in result.out
    assert ".avi" not in result.out
    assert "Lost - S01E01 - Pilot (1).mp4" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_format_specifiers(e2e_run, setup_test_files):
    setup_test_files("Ninja Turtles (1990).mkv")
    result = e2e_run(
        "--batch",
        "--movie_format={name} ({year}){extension}",
        "--movie_directory={name[0]}",
        "Ninja Turtles (1990).mkv",
    )
    expected = str(Path("/T/Teenage Mutant Ninja Turtles (1990).mkv"))
    assert result.code == 0
    assert expected in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_multiple_nested_directories(e2e_run, setup_test_files):
    setup_test_files("Ninja Turtles (1990).mkv")
    result = e2e_run(
        "--batch",
        "--movie_directory=1/2/3/",
        "Ninja Turtles (1990).mkv",
    )
    expected = str(Path("/1/2/3/Teenage Mutant Ninja Turtles (1990).mkv"))
    assert result.code == 0
    assert expected in result.out


@pytest.mark.omdb
@pytest.mark.usefixtures("setup_test_dir")
def test_format_id(e2e_run, setup_test_files):
    setup_test_files("aladdin.1992.avi")
    result = e2e_run(
        "--batch",
        "--media=movie",
        "--movie-api=omdb",
        "--movie-format='{name} ({id_imdb}).{extension}'",
        ".",
    )
    assert result.code == 0
    assert "Aladdin (tt0103639).avi" in result.out


@pytest.mark.tvdb
@pytest.mark.xfail(strict=False)
@pytest.mark.usefixtures("setup_test_dir")
def test_format_id__tvdb(e2e_run, setup_test_files):
    setup_test_files("archer.2009.s10e07.webrip.x264-lucidtv.mp4")
    result = e2e_run(
        "--batch",
        "--episode-api=tvdb",
        "--episode-format='{id_tvdb}.{season}x{episode}.{extension}'",
        ".",
    )
    assert result.code == 0
    assert "110381.10x7" in result.out


@pytest.mark.tvmaze
@pytest.mark.xfail(strict=False)
@pytest.mark.usefixtures("setup_test_dir")
def test_format_season0(e2e_run, setup_test_files):
    setup_test_files("south.park.s00e01.mp4")
    result = e2e_run(
        "--batch",
        "--episode-api=tvdb",
        "--episode-format='{series} {season:02}x{episode:02}.{extension}'",
        ".",
    )
    assert result.code == 0
    assert "South Park 00x01.mp4" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_replace_after(e2e_run, setup_test_files):
    setup_test_files("Pride & Prejudice 2005.ts")
    result = e2e_run("--batch", ".")
    assert result.code == 0
    assert "Pride and Prejudice (2005).ts" in result.out


@pytest.mark.usefixtures("setup_test_dir")
@pytest.mark.parametrize("container", SUBTITLE_CONTAINERS)
def test_subtitles(e2e_run, setup_test_files, container):
    setup_test_files(f"Saw (2004)/Eng{container}")
    result = e2e_run("--batch", "Saw (2004)")
    assert result.code == 0
    assert f"Saw (2004).en{container}" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_ambiguous_language_deletction(e2e_run, setup_test_files):
    setup_test_files(
        "Harry Potter and the Sorcerer's Stone 2001 Ultimate Extended Edition 1080p - KRaLiMaRKo.mkv"
    )
    result = e2e_run("--batch", ".")
    assert result.code == 0



import sys
from unittest.mock import patch

import pytest

from mnamer.argument import ArgLoader
from mnamer.setting_spec import SettingSpec
from mnamer.types import SettingType

pytestmark = pytest.mark.local


@pytest.mark.parametrize(
    "settings_type",
    (SettingType.DIRECTIVE, SettingType.PARAMETER, SettingType.POSITIONAL),
    ids=("directive", "parameter", "positional"),
)
def test_arg_loader__add_spec(settings_type: SettingType):
    arg_loader = ArgLoader()
    spec = SettingSpec(group=settings_type, flags=["-f"], help="foo")
    actions = getattr(arg_loader, f"_{settings_type.value}_group")
    assert len(actions._group_actions) == 0
    arg_loader += spec
    assert len(actions._group_actions) == 1


def test_arg_loader__add_spec_other():
    arg_loader = ArgLoader()
    spec = SettingSpec(group=SettingType.CONFIGURATION, flags=["-f"], help="foo")
    with pytest.raises(RuntimeError):
        arg_loader += spec


def test_arg_loader__format_help():
    arg_loader = ArgLoader()
    for spec in (
        SettingSpec(SettingType.POSITIONAL, flags=["--foo1"], help="foo1"),
        SettingSpec(SettingType.POSITIONAL, flags=["--foo2"], help="foo2"),
        SettingSpec(SettingType.POSITIONAL, flags=["--foo3"], help="foo3"),
        SettingSpec(SettingType.PARAMETER, flags=["--bar1"], help="bar1"),
        SettingSpec(SettingType.PARAMETER, flags=["--bar2"], help="bar2"),
        SettingSpec(SettingType.PARAMETER, flags=["--bar3"], help="bar3"),
        SettingSpec(SettingType.DIRECTIVE, flags=["--baz1"], help="baz1"),
        SettingSpec(SettingType.DIRECTIVE, flags=["--baz2"], help="baz2"),
        SettingSpec(SettingType.DIRECTIVE, flags=["--baz3"], help="baz3"),
    ):
        arg_loader._add_spec(spec)
    assert (
        arg_loader.format_help()
        == """
USAGE: mnamer [preferences] [directives] target [targets ...]

POSITIONAL:
  foo1
  foo2
  foo3

PARAMETERS:
  The following flags can be used to customize mnamer's behaviour. Their long
  forms may also be set in a '.mnamer-v2.json' config file, in which case cli
  arguments will take precedence.

  bar1
  bar2
  bar3

DIRECTIVES:
  Directives are one-off arguments that are used to perform secondary tasks
  like overriding media detection. They can't be used in '.mnamer-v2.json'.

  baz1
  baz2
  baz3

Visit https://github.com/jkwill87/mnamer for more information.
"""
    )


def test_arg_parser__load__valid_parameter():
    spec = SettingSpec(
        group=SettingType.PARAMETER, flags=["-f"], help="foo", typevar=int
    )
    arg_parser = ArgLoader(spec)
    with patch.object(sys, "argv", ["mnamer", "-f", "01"]):
        assert arg_parser.load() == {"f": 1}


def test_arg_parser__load__valid_directive():
    spec = SettingSpec(
        group=SettingType.DIRECTIVE, flags=["-f"], help="foo", typevar=int
    )
    arg_parser = ArgLoader(spec)
    with patch.object(sys, "argv", ["mnamer", "-f", "01"]):
        assert arg_parser.load() == {"f": 1}


def test_arg_parser__load__valid_positional():
    spec = SettingSpec(
        group=SettingType.POSITIONAL, flags=["f"], help="foo", typevar=int
    )
    arg_parser = ArgLoader(spec)
    with patch.object(sys, "argv", ["mnamer", "01"]):
        assert arg_parser.load() == {"f": 1}


def test_arg_parser__load__invalid_configuration():
    spec = SettingSpec(group=SettingType.CONFIGURATION, flags=["-f"], help="foo")
    arg_parser = ArgLoader(spec)
    with patch.object(sys, "argv", ["mnamer", "-f", "1"]):
        with pytest.raises(RuntimeError):
            arg_parser.load()


def test_arg_parser__missing_help():
    spec = SettingSpec(group=SettingType.DIRECTIVE, flags=["-f"])
    with pytest.raises(RuntimeError):
        ArgLoader(spec)



import pytest

from mnamer.setting_store import SettingStore
from mnamer.types import MediaType, ProviderType
from tests import DEFAULT_SETTINGS

pytestmark = pytest.mark.local


@pytest.mark.parametrize(
    "item", DEFAULT_SETTINGS.items(), ids=tuple(DEFAULT_SETTINGS.keys())
)
def test_as_dict(item):
    settings = SettingStore()
    k, v = item
    assert settings.as_dict()[k] == v


@pytest.mark.parametrize(
    "api", (ProviderType.TMDB, ProviderType.OMDB), ids=("TMDB", "OMDB")
)
def test_api_for__movie(api: ProviderType):
    settings = SettingStore(movie_api=api)
    assert settings.api_for(MediaType.MOVIE) is api


@pytest.mark.parametrize("api", ProviderType)
def test_api_key_for(api: ProviderType):
    settings = SettingStore()
    setattr(settings, f"api_key_{api.value}", "xxx")
    assert settings.api_key_for(api) == "xxx"



from unittest.mock import patch

import pytest

from mnamer.endpoints import omdb_search, omdb_title
from mnamer.exceptions import MnamerException, MnamerNotFoundException
from mnamer.providers import Omdb
from tests import JUNK_TEXT, MockRequestResponse

pytestmark = [
    pytest.mark.network,
    pytest.mark.omdb,
    pytest.mark.flaky(reruns=1),
]


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_omdb_title__title_id_xnor__title(mock_request):
    mock_response = MockRequestResponse(200, '{"key":"value"}')
    mock_request.return_value = mock_response
    omdb_title(Omdb.api_key, title="some title")


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_omdb_title__title_id_xnor__id(mock_request):
    mock_response = MockRequestResponse(200, '{"key":"value"}')
    mock_request.return_value = mock_response
    omdb_title(Omdb.api_key, id_imdb="tt123")


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_omdb_title__title_id_xnor__both(mock_request):
    mock_response = MockRequestResponse(200, '{"key":"value"}')
    mock_request.return_value = mock_response
    with pytest.raises(MnamerException):
        omdb_title(Omdb.api_key, title="some title", id_imdb="tt123")


@patch("mnamer.utils.requests_cache.CachedSession.request")
def test_omdb_title__title_id_xnor__neither(mock_request):
    mock_response = MockRequestResponse(200, '{"key":"value"}')
    mock_request.return_value = mock_response
    with pytest.raises(MnamerException):
        omdb_title(Omdb.api_key)


def test_omdb_title__media__movie():
    expected_top_level_keys = {
        "Actors",
        "Awards",
        "BoxOffice",
        "Country",
        "Director",
        "DVD",
        "Genre",
        "imdbID",
        "imdbRating",
        "imdbVotes",
        "Language",
        "Metascore",
        "Plot",
        "Poster",
        "Production",
        "Rated",
        "Ratings",
        "Released",
        "Response",
        "Runtime",
        "Title",
        "Type",
        "Website",
        "Writer",
        "Year",
    }
    result = omdb_title(Omdb.api_key, media="movie", title="ninja turtles")
    assert expected_top_level_keys.issuperset(set(result.keys()))
    assert result["Response"]
    assert result["Type"] == "movie"
    assert result["Title"] == "Teenage Mutant Ninja Turtles"


def test_omdb_title__media__series():
    expected_top_level_keys = {
        "Actors",
        "Awards",
        "Country",
        "Director",
        "Genre",
        "imdbID",
        "imdbRating",
        "imdbVotes",
        "Language",
        "Metascore",
        "Plot",
        "Poster",
        "Rated",
        "Ratings",
        "Released",
        "Response",
        "Runtime",
        "Title",
        "totalSeasons",
        "Type",
        "Writer",
        "Year",
    }

    result = omdb_title(Omdb.api_key, media="series", title="ninja turtles")
    assert set(result.keys()).issuperset(expected_top_level_keys)
    assert result["Response"]
    assert result["Type"] == "series"
    assert result["Title"] == "Teenage Mutant Ninja Turtles"


def test_omdb_title__api_key_fail():
    with pytest.raises(MnamerException):
        omdb_title(JUNK_TEXT, title="uhf", cache=False)


def test_omdb_title__id_imdb_fail():
    with pytest.raises(MnamerException):
        omdb_title(Omdb.api_key, "")


def test_omdb_title__not_found():
    with pytest.raises(MnamerNotFoundException):
        omdb_title(Omdb.api_key, "1" * 2)


def test_omdb_title__invalid_plot():
    with pytest.raises(MnamerException):
        omdb_title(Omdb.api_key, title="uhf", plot="medium")


def test_omdb_search__fields__top_level():
    expected_fields = {"Search", "Response", "totalResults"}
    result = omdb_search(Omdb.api_key, "ninja turtles")
    assert set(result.keys()) == expected_fields


def test_omdb_search__fields__search():
    expected_fields = {"Title", "Year", "imdbID", "Type", "Poster"}
    result = omdb_search(Omdb.api_key, "ninja turtles")["Search"][0]
    assert set(result.keys()) == expected_fields


def test_omdb_search__query__movie():
    result = omdb_search(Omdb.api_key, "ninja turtles", media="movie")
    assert all([entry["Type"] == "movie" for entry in result["Search"]])


def test_omdb_search__query__series():
    result = omdb_search(Omdb.api_key, "ninja turtles", media="series")
    assert all([entry["Type"] == "series" for entry in result["Search"]])


def test_omdb_search__api_key_fail():
    with pytest.raises(MnamerException):
        omdb_search(JUNK_TEXT, "ninja turtles", cache=False)


def test_omdb_search__query__fail():
    with pytest.raises(MnamerNotFoundException):
        omdb_search(Omdb.api_key, JUNK_TEXT, cache=False)


def test_omdb_search__year():
    result = omdb_search(Omdb.api_key, "ninja turtles", year=1987)
    assert "tt0131613" == result["Search"][0]["imdbID"]


def test_omdb_search__page_diff():
    p1 = omdb_search(Omdb.api_key, "Dogs", page=1)
    p2 = omdb_search(Omdb.api_key, "Dogs", page=2)
    assert p1 != p2


def test_omdb_search__page_out_of_bounds():
    with pytest.raises(MnamerException):
        omdb_search(Omdb.api_key, "Super Mario", page=101)



from unittest.mock import MagicMock, patch

import pytest

from mnamer.const import VERSION

pytestmark = [
    pytest.mark.e2e,
    pytest.mark.flaky(reruns=1),
]


@pytest.mark.parametrize("flag", ("-V", "--version"))
def test_version(flag: str, e2e_run):
    result = e2e_run(flag)
    assert result.code == 0
    assert result.out == f"mnamer version {VERSION}"


@patch("mnamer.frontends.clear_cache")
def test_directives__clear_cache(mock_clear_cache: MagicMock, e2e_run):
    result = e2e_run("--clear-cache")
    assert result.code == 0
    assert "cache cleared" in result.out
    mock_clear_cache.assert_called_once()


@pytest.mark.omdb
@pytest.mark.usefixtures("setup_test_dir")
def test_id__omdb(e2e_run, setup_test_files):
    setup_test_files("aladdin.1992.avi")
    result = e2e_run(
        "--batch",
        "--movie_api",
        "omdb",
        "--id-imdb",
        "tt5580390",
        ".",
    )
    assert "Shape of Water" in result.out


@pytest.mark.tmdb
@pytest.mark.usefixtures("setup_test_dir")
def test_id__tmdb(e2e_run, setup_test_files):
    setup_test_files("Ninja Turtles (1990).mkv")
    result = e2e_run("--batch", "--movie_api", "tmdb", "--id-tmdb", "475557", ".")
    assert result.code == 0
    assert "Joker" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_id__tvdb(e2e_run, setup_test_files):
    setup_test_files("game.of.thrones.01x05-eztv.mp4")
    result = e2e_run(
        "--batch",
        "--episode_api",
        "tvdb",
        "--id-tvdb",
        "79349",
        ".",
    )
    assert result.code == 0
    assert "Dexter" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_media__episode_override(e2e_run, setup_test_files):
    setup_test_files("aladdin.1992.avi")
    result = e2e_run("--batch", "--media", "episode", ".")
    assert result.code == 0
    assert "Processing Episode" in result.out


@pytest.mark.usefixtures("setup_test_dir")
def test_media__movie_override(e2e_run, setup_test_files):
    setup_test_files("s.w.a.t.2017.s02e01.mkv")
    result = e2e_run("--batch", "--media", "movie", ".")
    assert result.code == 0
    assert "Processing Movie" in result.out


def test_test(e2e_run):
    result = e2e_run("--batch", "--test", ".")
    assert result.code == 0
    assert "testing mode" in result.out



import datetime as dt
from pathlib import Path

import pytest

from mnamer.metadata import MetadataEpisode, MetadataMovie
from mnamer.setting_store import SettingStore
from mnamer.target import Target
from mnamer.types import MediaType

pytestmark = pytest.mark.local


def test_parse__media__movie():
    target = Target(Path("ninja turtles (1990).mkv"), SettingStore())
    assert target.metadata.to_media_type() is MediaType.MOVIE


def test_parse__media__episode():
    target = Target(Path("ninja turtles s01e01.mkv"), SettingStore())
    assert target.metadata.to_media_type() is MediaType.EPISODE


def test_parse__quality():
    file_path = Path("ninja.turtles.s01e04.1080p.ac3.rargb.sample.mkv")
    target = Target(file_path, SettingStore())
    assert target.metadata.quality == "1080p dolby digital"


def test_parse__group():
    file_path = Path("ninja.turtles.s01e04.1080p.ac3.rargb.sample.mkv")
    target = Target(file_path, SettingStore())
    assert target.metadata.group == "RARGB"


def test_parse__container():
    file_path = Path("ninja.turtles.s01e04.1080p.ac3.rargb.sample.mp4")
    target = Target(file_path, SettingStore())
    assert target.metadata.container == ".mp4"


def test_parse__date():
    file_path = Path("the.colbert.show.2010.10.01.avi")
    target = Target(file_path, SettingStore())
    assert isinstance(target.metadata, MetadataEpisode)
    assert target.metadata.date == dt.date(2010, 10, 1)


def test_parse__episode():
    file_path = Path("ninja.turtles.s01e04.1080p.ac3.rargb.sample.mp4")
    target = Target(file_path, SettingStore())
    assert isinstance(target.metadata, MetadataEpisode)
    assert target.metadata.episode == 4


def test_parse__season():
    file_path = Path("ninja.turtles.s01e04.1080p.ac3.rargb.sample.mp4")
    target = Target(file_path, SettingStore())
    assert isinstance(target.metadata, MetadataEpisode)
    assert target.metadata.season == 1


def test_parse__series():
    file_path = Path("ninja.turtles.s01e04.1080p.ac3.rargb.sample.mp4")
    target = Target(file_path, SettingStore())
    assert isinstance(target.metadata, MetadataEpisode)
    assert target.metadata.series == "Ninja Turtles"


def test_parse__year():
    file_path = Path("the.goonies.1985")
    target = Target(file_path, SettingStore())
    assert isinstance(target.metadata, MetadataMovie)
    assert target.metadata.year == 1985


def testparse__name():
    file_path = Path("the.goonies.1985")
    target = Target(file_path, SettingStore())
    assert isinstance(target.metadata, MetadataMovie)
    assert target.metadata.name == "The Goonies"


@pytest.mark.parametrize("media", MediaType)
def test_media__override(media: MediaType):
    target = Target(Path(), SettingStore(media=media))
    assert target.metadata.to_media_type() == media


def test_directory__movie():
    movie_path = Path("/some/movie/path").absolute()
    target = Target(
        Path(), SettingStore(media=MediaType.MOVIE, movie_directory=movie_path)
    )
    assert target.directory == movie_path


def test_directory__episode():
    episode_path = Path("/some/episode/path").absolute()
    target = Target(
        Path(),
        SettingStore(media=MediaType.EPISODE, episode_directory=episode_path),
    )
    assert target.directory == episode_path


def test_ambiguous_subtitle_language():
    target = Target(
        Path("Subs/Nancy.Drew.S01E01.WEBRip.x264-ION10.srt"), SettingStore()
    )
    assert target.metadata.language is None


def test_destination__simple():
    pass  # TODO


def test_query():
    pass  # TODO


def test_relocate():
    pass  # TODO



import pytest

from mnamer.exceptions import MnamerNotFoundException
from mnamer.metadata import MetadataEpisode
from mnamer.providers import TvMaze
from tests import EPISODE_META, JUNK_TEXT, TEST_DATE

pytestmark = [
    pytest.mark.network,
    pytest.mark.tvmaze,
    pytest.mark.flaky(reruns=1),
]


@pytest.fixture(scope="session")
def provider():
    return TvMaze()


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id_tvmaze_and_season_and_episode(meta: dict, provider: TvMaze):
    query = MetadataEpisode(
        id_tvmaze=meta["id_tvmaze"], season=meta["season"], episode=meta["episode"]
    )
    results = list(provider.search(query))
    assert results
    for result in results:
        assert result.title == meta["title"]


def test_search_id_tvmaze_and_season_and_episode__no_hits(provider: TvMaze):
    query = MetadataEpisode(id_tvmaze=JUNK_TEXT, season=1, episode=1)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id_tvmaze_and_date(meta: dict, provider: TvMaze):
    query = MetadataEpisode(id_tvmaze=meta["id_tvmaze"], date=meta["date"])
    results = list(provider.search(query))
    assert results
    for result in results:
        assert result.title == meta["title"]


def test_search_id_tvmaze_and_date__no_hits(provider: TvMaze):
    query = MetadataEpisode(id_tvmaze=JUNK_TEXT, date=TEST_DATE)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id_tvdb_and_date(meta, provider: TvMaze):
    query = MetadataEpisode(id_tvdb=meta["id_tvdb"], date=meta["date"])
    results = list(provider.search(query))
    assert results
    for result in results:
        assert result.title == meta["title"]


def test_search_id_tvdb_and_date__no_hits(provider: TvMaze):
    query = MetadataEpisode(id_tvdb=JUNK_TEXT, date=TEST_DATE)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id_tvmaze_and_season(meta: dict, provider: TvMaze):
    query = MetadataEpisode(id_tvmaze=meta["id_tvmaze"], season=meta["season"])
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)
    assert all(result.season == meta["season"] for result in results)


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id_tvmaze_and_episode(meta: dict, provider: TvMaze):
    query = MetadataEpisode(id_tvmaze=meta["id_tvmaze"], episode=meta["episode"])
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)
    assert all(result.episode == meta["episode"] for result in results)


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id_tvmaze(meta: dict, provider: TvMaze):
    query = MetadataEpisode(id_tvmaze=meta["id_tvmaze"])
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)


def test_search_id_tvmaze__no_hits(provider: TvMaze):
    query = MetadataEpisode(id_tvmaze=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_id_tvdb(meta: dict, provider: TvMaze):
    query = MetadataEpisode(id_tvdb=meta["id_tvdb"])
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)


def test_search_id_tvdb__no_hits(provider: TvMaze):
    query = MetadataEpisode(id_tvdb=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_series_and_season_and_episode(meta: dict, provider: TvMaze):
    query = MetadataEpisode(
        series=meta["series"], season=meta["season"], episode=meta["episode"]
    )
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)


def test_search_series_and_season_and_episode__no_hits(provider: TvMaze):
    query = MetadataEpisode(series=JUNK_TEXT, season=1, episode=1)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_series_and_season(meta: dict, provider: TvMaze):
    query = MetadataEpisode(series=meta["series"], season=meta["season"])
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)


def test_search_series_and_season__no_hits(provider):
    query = MetadataEpisode(series=JUNK_TEXT, season=1)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_series_and_episode(meta: dict, provider: TvMaze):
    query = MetadataEpisode(series=meta["series"], episode=meta["episode"])
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)


def test_search_series_and_episode__no_hits(provider: TvMaze):
    query = MetadataEpisode(series=JUNK_TEXT, episode=1)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


@pytest.mark.parametrize("meta", EPISODE_META.values(), ids=list(EPISODE_META))
def test_search_series(meta: dict, provider: TvMaze):
    query = MetadataEpisode(series=meta["series"])
    results = list(provider.search(query))
    assert results
    assert any(result.title == meta["title"] for result in results)


def test_search_series__no_hits(provider: TvMaze):
    query = MetadataEpisode(series=JUNK_TEXT)
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))


def test_search__no_hits(provider: TvMaze):
    query = MetadataEpisode()
    with pytest.raises(MnamerNotFoundException):
        next(provider.search(query))
